"""
====================================
Streamline length and size reduction
====================================

This example shows how to calculate the lengths of a set of streamlines and
also how to compress the streamlines without considerably reducing their
lengths or overall shape.

A streamline in DIPY_ is represented as a numpy array of size
:math:`(N \times 3)` where each row of the array represents a 3D point of the
streamline. A set of streamlines is represented with a list of
numpy arrays of size :math:`(N_i \times 3)` for :math:`i=1:M` where $M$ is the
number of streamlines in the set.
"""

import matplotlib.pyplot as plt
import numpy as np

from dipy.tracking.distances import approx_polygon_track
from dipy.tracking.streamline import set_number_of_points
from dipy.tracking.utils import length
from dipy.viz import actor, window

###############################################################################
# Let's first create a simple simulation of a bundle of streamlines using
# a cosine function.


def simulated_bundles(no_streamlines=50, n_pts=100):
    rng = np.random.default_rng()

    t = np.linspace(-10, 10, n_pts)

    bundle = []
    for i in np.linspace(3, 5, no_streamlines):
        pts = np.vstack((np.cos(2 * t / np.pi), np.zeros(t.shape) + i, t)).T
        bundle.append(pts)

    start = rng.integers(10, 30, no_streamlines)
    end = rng.integers(60, 100, no_streamlines)

    bundle = [
        10 * streamline[start[i] : end[i]] for (i, streamline) in enumerate(bundle)
    ]
    bundle = [np.ascontiguousarray(streamline) for streamline in bundle]

    return bundle


bundle = simulated_bundles()

print(f"This bundle has {len(bundle)} streamlines")

###############################################################################
# Using the ``length`` function we can retrieve the lengths of each streamline.
# Below we show the histogram of the lengths of the streamlines.

lengths = list(length(bundle))

fig_hist, ax = plt.subplots(1)
ax.hist(lengths, color="burlywood")
ax.set_xlabel("Length")
ax.set_ylabel("Count")
# plt.show()
plt.legend()
plt.savefig("length_histogram.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Histogram of lengths of the streamlines
#
#
# ``Length`` will return the length in the units of the coordinate system that
# streamlines are currently. So, if the streamlines are in world coordinates
# then the lengths will be in millimeters (mm). If the streamlines are for
# example in native image coordinates of voxel size 2mm isotropic then you
# will need to multiply the lengths by 2 if you want them to correspond to mm.
# In this example we process simulated data without units, however this
# information is good to have in mind when you calculate lengths with real
# data.
#
# Next, let's find the number of points that each streamline has.

n_pts = [len(streamline) for streamline in bundle]

###############################################################################
# Often, streamlines are represented with more points than what is actually
# necessary for specific applications. Also, sometimes every streamline has a
# different number of points, which could be a problem for some algorithms.
# The function ``set_number_of_points`` can be used to set the number of
# points of a streamline at a specific number and at the same time enforce
# that all the segments of the streamline will have equal length.

bundle_downsampled = set_number_of_points(bundle, nb_points=12)
n_pts_ds = [len(s) for s in bundle_downsampled]

###############################################################################
# Alternatively, the function ``approx_polygon_track`` allows reducing the
# number of points so that there are more points in curvy regions and less
# points in less curvy regions. In contrast with ``set_number_of_points`` it
# does not enforce that segments should be of equal size.

bundle_downsampled2 = [approx_polygon_track(s, 0.25) for s in bundle]
n_pts_ds2 = [len(streamline) for streamline in bundle_downsampled2]

###############################################################################
# Both, ``set_number_of_points`` and ``approx_polygon_track`` can be thought as
# methods for lossy compression of streamlines.

# Enables/disables interactive visualization
interactive = False

scene = window.Scene()
scene.SetBackground(*window.colors.white)
bundle_actor = actor.streamtube(bundle, colors=window.colors.red, linewidth=0.3)

scene.add(bundle_actor)

bundle_actor2 = actor.streamtube(
    bundle_downsampled, colors=window.colors.red, linewidth=0.3
)
bundle_actor2.SetPosition(0, 40, 0)

bundle_actor3 = actor.streamtube(
    bundle_downsampled2, colors=window.colors.red, linewidth=0.3
)
bundle_actor3.SetPosition(0, 80, 0)

scene.add(bundle_actor2)
scene.add(bundle_actor3)

scene.set_camera(position=(0, 0, 0), focal_point=(30, 0, 0))
window.record(scene=scene, out_path="simulated_cosine_bundle.png", size=(900, 900))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Initial bundle (down), downsampled at 12 equidistant points (middle),
# downsampled with points that are not equidistant (up).
#
#
# From the figure above we can see that all 3 bundles look quite similar.
# However, when we plot the histogram of the number of points used for each
# streamline, it becomes obvious that we have managed to reduce in a great
# amount the size of the initial dataset.

fig_hist, ax = plt.subplots(1)
ax.hist(n_pts, color="r", histtype="step", label="initial")
ax.hist(n_pts_ds, color="g", histtype="step", label="set_number_of_points (12)")
ax.hist(n_pts_ds2, color="b", histtype="step", label="approx_polygon_track (0.25)")
ax.set_xlabel("Number of points")
ax.set_ylabel("Count")

# plt.show()
plt.legend()
plt.savefig("n_pts_histogram.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Histogram of the number of points of the streamlines.
#
#
# Finally, we can also show that the lengths of the streamlines haven't changed
# considerably after applying the two methods of downsampling.

lengths_downsampled = list(length(bundle_downsampled))
lengths_downsampled2 = list(length(bundle_downsampled2))

fig, ax = plt.subplots(1)
ax.plot(lengths, color="r", label="initial")
ax.plot(lengths_downsampled, color="g", label="set_number_of_points (12)")
ax.plot(lengths_downsampled2, color="b", label="approx_polygon_track (0.25)")
ax.set_xlabel("Streamline ID")
ax.set_ylabel("Length")

# plt.show()
plt.legend()
plt.savefig("lengths_plots.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Lengths of each streamline for every one of the 3 bundles.



"""
=============================================================================
Using the free water elimination model to remove DTI free water contamination
=============================================================================

As shown previously (see
:ref:`sphx_glr_examples_built_reconstruction_reconst_dti.py`), the diffusion
tensor model is a simple way to characterize diffusion anisotropy. However,
in regions near the ventricles and parenchyma, anisotropy can be
underestimated by partial volume effects of the cerebral spinal fluid (CSF).
This free water contamination can particularly corrupt Diffusion Tensor
Imaging analysis of microstructural changes when different groups of subjects
show different brain morphology (e.g. brain ventricle enlargement associated
with brain tissue atrophy that occurs in several brain pathologies and aging).

A way to remove this free water influences is to expand the DTI model to take
into account an extra compartment representing the contributions of free water
diffusion :footcite:p:`Pasternak2009`. The expression of the expanded DTI model
is shown below:

.. math::

    S(\\mathbf{g}, b) = S_0(1-f)e^{-b\\mathbf{g}^T \\mathbf{D}
    \\mathbf{g}}+S_0fe^{-b D_{iso}}

where $\\mathbf{g}$ and $b$ are diffusion gradient direction and weighted (more
information see :ref:`sphx_glr_examples_built_reconstruction_reconst_dti.py`),
$S(\\mathbf{g}, b)$ is thebdiffusion-weighted signal measured, $S_0$ is the
signal in a measurement with no diffusion weighting, $\\mathbf{D}$ is the
diffusion tensor, $f$ the volume fraction of the free water component, and
$D_{iso}$ is the isotropic value of the free water diffusion (normally set to
$3.0 \times 10^{-3} mm^{2}s^{-1}$).

In this example, we show how to process a diffusion weighting dataset using an
adapted version of the free water elimination proposed by :footcite:p:`Hoy2014`.

The full details of Dipy's free water DTI implementation was published in
:footcite:p:`NetoHenriques2017`. Please cite this work if you use this
algorithm.

Let's start by importing the relevant modules:
"""

import os.path as op

import matplotlib.pyplot as plt
import nibabel as nib
import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import fetch_hbn
import dipy.reconst.dti as dti
import dipy.reconst.fwdti as fwdti

###############################################################################
# Without spatial constrains the free water elimination model cannot be solved
# in data acquired from one non-zero b-value :footcite:p:`Hoy2014`. Therefore,
# here we download a dataset that was acquired with multiple b-values.

data_path = fetch_hbn(["NDARAA948VFH"])[1]
dwi_path = op.join(
    data_path, "derivatives", "qsiprep", "sub-NDARAA948VFH", "ses-HBNsiteRU", "dwi"
)

img = nib.load(
    op.join(
        dwi_path,
        "sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi.nii.gz",
    )
)

gtab = gradient_table(
    op.join(
        dwi_path,
        "sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi.bval",
    ),
    bvecs=op.join(
        dwi_path,
        "sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi.bvec",
    ),
)

data = np.asarray(img.dataobj)

###############################################################################
# The free water DTI model can take some minutes to process the full data set.
# Thus, we use a brain mask that was calculated during pre-processing, to
# remove the background of the image and avoid unnecessary calculations.

mask_img = nib.load(
    op.join(
        dwi_path,
        "sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-brain_mask.nii.gz",
    )
)

###############################################################################
# Moreover, for illustration purposes we process only one slice of the data.

mask = mask_img.get_fdata()

data_small = data[:, :, 50:51]
mask_small = mask[:, :, 50:51]

###############################################################################
# The free water elimination model fit can then be initialized by instantiating
# a FreeWaterTensorModel class object:

fwdtimodel = fwdti.FreeWaterTensorModel(gtab)

###############################################################################
# The data can then be fitted using the ``fit`` function of the defined model
# object:

fwdtifit = fwdtimodel.fit(data_small, mask=mask_small)

###############################################################################
# This 2-steps procedure will create a FreeWaterTensorFit object which contains
# all the diffusion tensor statistics free for free water contamination. Below
# we extract the fractional anisotropy (FA) and the mean diffusivity (MD) of
# the free water diffusion tensor.

FA = fwdtifit.fa
MD = fwdtifit.md

###############################################################################
# For comparison we also compute the same standard measures processed by the
# standard DTI model

dtimodel = dti.TensorModel(gtab)

dtifit = dtimodel.fit(data_small, mask=mask_small)

dti_FA = dtifit.fa
dti_MD = dtifit.md

###############################################################################
# Below the FA values for both free water elimination DTI model and standard
# DTI model are plotted in panels A and B, while the respective MD values are
# plotted in panels D and E. For a better visualization of the effect of the
# free water correction, the differences between these two metrics are shown
# in panels C and E. In addition to the standard diffusion statistics, the
# estimated volume fraction of the free water contamination is shown on
# panel G.

fig1, ax = plt.subplots(2, 4, figsize=(12, 6), subplot_kw={"xticks": [], "yticks": []})

fig1.subplots_adjust(hspace=0.3, wspace=0.05)
ax.flat[0].imshow(FA[:, :, 0].T, origin="lower", cmap="gray", vmin=0, vmax=1)
ax.flat[0].set_title("A) fwDTI FA")
ax.flat[1].imshow(dti_FA[:, :, 0].T, origin="lower", cmap="gray", vmin=0, vmax=1)
ax.flat[1].set_title("B) standard DTI FA")

FAdiff = abs(FA[:, :, 0] - dti_FA[:, :, 0])
ax.flat[2].imshow(FAdiff.T, cmap="gray", origin="lower", vmin=0, vmax=1)
ax.flat[2].set_title("C) FA difference")

ax.flat[3].axis("off")

ax.flat[4].imshow(MD[:, :, 0].T, origin="lower", cmap="gray", vmin=0, vmax=2.5e-3)
ax.flat[4].set_title("D) fwDTI MD")
ax.flat[5].imshow(dti_MD[:, :, 0].T, origin="lower", cmap="gray", vmin=0, vmax=2.5e-3)
ax.flat[5].set_title("E) standard DTI MD")

MDdiff = abs(MD[:, :, 0] - dti_MD[:, :, 0])
ax.flat[6].imshow(MDdiff.T, origin="lower", cmap="gray", vmin=0, vmax=2.5e-3)
ax.flat[6].set_title("F) MD difference")

F = fwdtifit.f

ax.flat[7].imshow(F[:, :, 0].T, origin="lower", cmap="gray", vmin=0, vmax=1)
ax.flat[7].set_title("G) free water volume")

plt.show()
fig1.savefig("In_vivo_free_water_DTI_and_standard_DTI_measures.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# In vivo diffusion measures obtain from the free water DTI and standard
# DTI. The values of Fractional Anisotropy for the free water DTI model and
# standard DTI model and their difference are shown in the upper panels (A-C),
# while respective MD values are shown in the lower panels (D-F). In addition
# the free water volume fraction estimated from the fwDTI model is shown in
# panel G.
#
#
# From the figure, one can observe that the free water elimination model
# produces in general higher values of FA and lower values of MD than the
# standard DTI model. These differences in FA and MD estimation are expected
# due to the suppression of the free water isotropic diffusion components.
# Unexpected high amplitudes of FA are however observed in the periventricular
# gray matter. This is a known artefact of regions associated to voxels with
# high water volume fraction (i.e. voxels containing basically CSF). We are
# able to remove this problematic voxels by excluding all FA values
# associated with measured volume fractions above a reasonable threshold
# of 0.7:

FA[F > 0.7] = 0
dti_FA[F > 0.7] = 0

###############################################################################
# Above we reproduce the plots of the in vivo FA from the two DTI fits and
# where we can see that the inflated FA values were practically removed:

fig1, ax = plt.subplots(1, 3, figsize=(9, 3), subplot_kw={"xticks": [], "yticks": []})

fig1.subplots_adjust(hspace=0.3, wspace=0.05)
ax.flat[0].imshow(FA[:, :, 0].T, origin="lower", cmap="gray", vmin=0, vmax=1)
ax.flat[0].set_title("A) fwDTI FA")
ax.flat[1].imshow(dti_FA[:, :, 0].T, origin="lower", cmap="gray", vmin=0, vmax=1)
ax.flat[1].set_title("B) standard DTI FA")

FAdiff = abs(FA[:, :, 0] - dti_FA[:, :, 0])
ax.flat[2].imshow(FAdiff.T, cmap="gray", origin="lower", vmin=0, vmax=1)
ax.flat[2].set_title("C) FA difference")

plt.show()
fig1.savefig("In_vivo_free_water_DTI_and_standard_DTI_corrected.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# In vivo FA measures obtain from the free water DTI (A) and standard
# DTI (B) and their difference (C). Problematic inflated FA values of the
# images were removed by dismissing voxels above a volume fraction threshold
# of 0.7.
#
#
# References
# ----------
#
# .. footbibliography::
#



import io
from os.path import splitext
import re
import warnings

import numpy as np


def read_bvals_bvecs(fbvals, fbvecs):
    """Read b-values and b-vectors from disk.

    Parameters
    ----------
    fbvals : str
       Full path to file with b-values. None to not read bvals.
    fbvecs : str
       Full path of file with b-vectors. None to not read bvecs.

    Returns
    -------
    bvals : array, (N,) or None
    bvecs : array, (N, 3) or None

    Notes
    -----
    Files can be either '.bvals'/'.bvecs' or '.txt' or '.npy' (containing
    arrays stored with the appropriate values).

    """
    # Loop over the provided inputs, reading each one in turn and adding them
    # to this list:
    vals = []
    for this_fname in [fbvals, fbvecs]:
        # If the input was None or empty string, we don't read anything and
        # move on:
        if this_fname is None or not this_fname:
            vals.append(None)
            continue

        if not isinstance(this_fname, str):
            raise ValueError("String with full path to file is required")

        base, ext = splitext(this_fname)
        if ext in [
            ".bvals",
            ".bval",
            ".bvecs",
            ".bvec",
            ".txt",
            ".eddy_rotated_bvecs",
            "",
        ]:
            with open(this_fname, "r") as f:
                content = f.read()

            munged_content = io.StringIO(re.sub(r"(\t|,)", " ", content))
            vals.append(np.squeeze(np.loadtxt(munged_content)))
        elif ext == ".npy":
            vals.append(np.squeeze(np.load(this_fname)))
        else:
            e_s = f"File type {ext} is not recognized"
            raise ValueError(e_s)

    # Once out of the loop, unpack them:
    bvals, bvecs = vals[0], vals[1]

    # If bvecs is None, you can just return now w/o making more checks:
    if bvecs is None:
        return bvals, bvecs

    if 3 not in bvecs.shape:
        raise OSError("bvec file should have three rows")
    if bvecs.ndim != 2:
        bvecs = bvecs[None, ...]
        bvals = bvals[None, ...]
        msg = "Detected only 1 direction on your bvec file. For diffusion "
        msg += "dataset, it is recommended to have at least 3 directions."
        msg += "You may have problems during the reconstruction step."
        warnings.warn(msg, stacklevel=2)
    if bvecs.shape[1] != 3:
        bvecs = bvecs.T

    # If bvals is None, you don't need to check that they have the same shape:
    if bvals is None:
        return bvals, bvecs

    if len(bvals.shape) > 1:
        raise OSError("bval file should have one row")

    if bvals.shape[0] != bvecs.shape[0]:
        raise OSError("b-values and b-vectors shapes do not correspond")

    return bvals, bvecs



import warnings

import numpy as np
import numpy.testing as npt
import pytest

import dipy.core.gradients as grad
import dipy.core.optimize as opt
import dipy.data as dpd
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti_data
import dipy.reconst.cross_validation as xval
import dipy.reconst.sfm as sfm
import dipy.sims.voxel as sims
from dipy.utils.optpkg import optional_package

warnings.filterwarnings("ignore")

sklearn, has_sklearn, _ = optional_package("sklearn")
needs_sklearn = pytest.mark.skipif(not has_sklearn, reason="Requires sklearn")


def test_design_matrix():
    data, gtab = dpd.dsi_voxels()
    sphere = dpd.get_sphere()
    # Make it with NNLS, so that it gets tested regardless of sklearn
    sparse_fascicle_model = sfm.SparseFascicleModel(gtab, sphere=sphere, solver="NNLS")
    npt.assert_equal(
        sparse_fascicle_model.design_matrix.shape,
        (np.sum(~gtab.b0s_mask), sphere.vertices.shape[0]),
    )


@needs_sklearn
def test_sfm():
    fdata, fbvals, fbvecs = dpd.get_fnames()
    data = load_nifti_data(fdata)
    gtab = grad.gradient_table(fbvals, bvecs=fbvecs)
    for n_procs in [1, 2]:
        for iso in [sfm.ExponentialIsotropicModel, None]:
            sfmodel = sfm.SparseFascicleModel(gtab, isotropic=iso)
            sffit1 = sfmodel.fit(data[0, 0, 0], num_processes=n_procs)
            sphere = dpd.get_sphere()
            odf1 = sffit1.odf(sphere)
            pred1 = sffit1.predict(gtab=gtab)
            mask = np.ones(data.shape[:-1])
            sffit2 = sfmodel.fit(data, mask=mask, num_processes=n_procs)
            pred2 = sffit2.predict(gtab=gtab)
            odf2 = sffit2.odf(sphere)
            sffit3 = sfmodel.fit(data, num_processes=n_procs)
            pred3 = sffit3.predict(gtab=gtab)
            odf3 = sffit3.odf(sphere)
            npt.assert_almost_equal(pred3, pred2, decimal=2)
            npt.assert_almost_equal(pred3[0, 0, 0], pred1, decimal=2)
            npt.assert_almost_equal(odf3[0, 0, 0], odf1, decimal=2)
            npt.assert_almost_equal(odf3[0, 0, 0], odf2[0, 0, 0], decimal=2)
            # Fit zeros and you will get back zeros
            npt.assert_almost_equal(
                sfmodel.fit(np.zeros(data[0, 0, 0].shape), num_processes=n_procs).beta,
                np.zeros(sfmodel.design_matrix[0].shape[-1]),
            )


@needs_sklearn
def test_predict():
    SNR = 1000
    S0 = 100
    _, fbvals, fbvecs = dpd.get_fnames(name="small_64D")
    bvals, bvecs = read_bvals_bvecs(fbvals, fbvecs)
    gtab = grad.gradient_table(bvals, bvecs=bvecs)
    mevals = np.array(([0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003]))
    angles = [(0, 0), (60, 0)]
    S, sticks = sims.multi_tensor(
        gtab, mevals, S0=S0, angles=angles, fractions=[10, 90], snr=SNR
    )

    sfmodel = sfm.SparseFascicleModel(gtab, response=[0.0015, 0.0003, 0.0003])
    sffit = sfmodel.fit(S)
    pred = sffit.predict()
    npt.assert_(xval.coeff_of_determination(pred, S) > 97)

    # Should be possible to predict using a different gtab:
    new_gtab = grad.gradient_table(bvals[::2], bvecs=bvecs[::2])
    new_pred = sffit.predict(gtab=new_gtab)
    npt.assert_(xval.coeff_of_determination(new_pred, S[::2]) > 97)

    # Should be possible to predict for a single direction:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        new_gtab = grad.gradient_table(bvals[1][None], bvecs=bvecs[1][None, :])
    new_pred = sffit.predict(gtab=new_gtab)

    # Fitting and predicting with a volume of data:
    fdata, fbval, fbvec = dpd.get_fnames(name="small_25")
    gtab = grad.gradient_table(fbval, bvecs=fbvec)
    data = load_nifti_data(fdata)
    sfmodel = sfm.SparseFascicleModel(gtab, response=[0.0015, 0.0003, 0.0003])
    sffit = sfmodel.fit(data)
    pred = sffit.predict()

    # Should be possible to predict using a different gtab:
    new_gtab = grad.gradient_table(bvals[::2], bvecs=bvecs[::2])
    new_pred = sffit.predict(gtab=new_gtab)
    npt.assert_equal(
        new_pred.shape,
        data.shape[:-1] + bvals[::2].shape,
    )

    # Should be possible to predict for a single direction:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        new_gtab = grad.gradient_table(bvals[1][None], bvecs=bvecs[1][None, :])
    new_pred = sffit.predict(gtab=new_gtab)
    npt.assert_equal(new_pred.shape, data.shape[:-1])

    # Fitting and predicting with masked data:
    mask = np.zeros(data.shape[:3])
    mask[2:5, 2:5, :] = 1
    sffit = sfmodel.fit(data, mask=mask)
    pred = sffit.predict()
    npt.assert_equal(pred.shape, data.shape)

    # Should be possible to predict using a different gtab:
    new_gtab = grad.gradient_table(bvals[::2], bvecs=bvecs[::2])
    new_pred = sffit.predict(gtab=new_gtab)
    npt.assert_equal(
        new_pred.shape,
        data.shape[:-1] + bvals[::2].shape,
    )
    npt.assert_equal(new_pred[0, 0, 0], 0)

    # Should be possible to predict for a single direction:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        new_gtab = grad.gradient_table(bvals[1][None], bvecs=bvecs[1][None, :])
    new_pred = sffit.predict(gtab=new_gtab)
    npt.assert_equal(new_pred.shape, data.shape[:-1])
    npt.assert_equal(new_pred[0, 0, 0], 0)


def test_sfm_background():
    fdata, fbvals, fbvecs = dpd.get_fnames()
    data = load_nifti_data(fdata)
    gtab = grad.gradient_table(fbvals, bvecs=fbvecs)
    to_fit = data[0, 0, 0]
    to_fit[gtab.b0s_mask] = 0
    sfmodel = sfm.SparseFascicleModel(gtab, solver="NNLS")
    sffit = sfmodel.fit(to_fit)
    npt.assert_equal(sffit.beta, np.zeros_like(sffit.beta))


def test_sfm_stick():
    fdata, fbvals, fbvecs = dpd.get_fnames()
    data = load_nifti_data(fdata)
    gtab = grad.gradient_table(fbvals, bvecs=fbvecs)
    sfmodel = sfm.SparseFascicleModel(gtab, solver="NNLS", response=[0.001, 0, 0])
    sffit1 = sfmodel.fit(data[0, 0, 0])
    sphere = dpd.get_sphere()
    sffit1.odf(sphere)
    sffit1.predict(gtab=gtab)

    SNR = 1000
    S0 = 100
    mevals = np.array(([0.001, 0, 0], [0.001, 0, 0]))
    angles = [(0, 0), (60, 0)]
    S, sticks = sims.multi_tensor(
        gtab, mevals, S0=S0, angles=angles, fractions=[50, 50], snr=SNR
    )

    sfmodel = sfm.SparseFascicleModel(gtab, solver="NNLS", response=[0.001, 0, 0])
    sffit = sfmodel.fit(S)
    pred = sffit.predict()
    npt.assert_(xval.coeff_of_determination(pred, S) > 96)


@needs_sklearn
def test_sfm_sklearnlinearsolver():
    class SillySolver(opt.SKLearnLinearSolver):
        def fit(self, X, y):
            self.coef_ = np.ones(X.shape[-1])

    class EvenSillierSolver:
        def fit(self, X, y):
            self.coef_ = np.ones(X.shape[-1])

    fdata, fbvals, fbvecs = dpd.get_fnames()
    gtab = grad.gradient_table(fbvals, bvecs=fbvecs)
    sfmodel = sfm.SparseFascicleModel(gtab, solver=SillySolver())

    npt.assert_(isinstance(sfmodel.solver, SillySolver))
    npt.assert_raises(
        ValueError, sfm.SparseFascicleModel, gtab, solver=EvenSillierSolver()
    )


@needs_sklearn
def test_exponential_iso():
    fdata, fbvals, fbvecs = dpd.get_fnames()
    data_dti = load_nifti_data(fdata)
    gtab_dti = grad.gradient_table(fbvals, bvecs=fbvecs)
    data_multi, gtab_multi = dpd.dsi_deconv_voxels()

    for data, gtab in zip([data_dti, data_multi], [gtab_dti, gtab_multi]):
        sfmodel = sfm.SparseFascicleModel(gtab, isotropic=sfm.ExponentialIsotropicModel)

        sffit1 = sfmodel.fit(data[0, 0, 0])
        sphere = dpd.get_sphere()
        odf = sffit1.odf(sphere)
        pred = sffit1.predict(gtab=gtab)
        npt.assert_equal(pred.shape, data[0, 0, 0].shape)
        npt.assert_equal(odf.shape, data[0, 0, 0].shape[:-1] + (sphere.x.shape[0],))

        sffit2 = sfmodel.fit(data)
        sphere = dpd.get_sphere()
        odf = sffit2.odf(sphere)
        pred = sffit2.predict(gtab=gtab)
        npt.assert_equal(pred.shape, data.shape)
        npt.assert_equal(odf.shape, data.shape[:-1] + (sphere.x.shape[0],))

        mask = np.zeros(data.shape[:3])
        mask[2:5, 2:5, :] = 1
        sffit3 = sfmodel.fit(data, mask=mask)
        sphere = dpd.get_sphere()
        odf = sffit3.odf(sphere)
        pred = sffit3.predict(gtab=gtab)
        npt.assert_equal(pred.shape, data.shape)
        npt.assert_equal(odf.shape, data.shape[:-1] + (sphere.x.shape[0],))

        SNR = 1000
        S0 = 100
        mevals = np.array(([0.0015, 0.0005, 0.0005], [0.0015, 0.0005, 0.0005]))
        angles = [(0, 0), (60, 0)]
        S, sticks = sims.multi_tensor(
            gtab, mevals, S0=S0, angles=angles, fractions=[50, 50], snr=SNR
        )
        sffit = sfmodel.fit(S)
        pred = sffit.predict()
        npt.assert_(xval.coeff_of_determination(pred, S) > 96)



import numpy as np

from dipy.reconst.base import ReconstFit, ReconstModel
from dipy.testing.decorators import warning_for_keywords

# Classes OdfModel and OdfFit are using API ReconstModel and ReconstFit from
# .base


class OdfModel(ReconstModel):
    """An abstract class to be sub-classed by specific odf models

    All odf models should provide a fit method which may take data as it's
    first and only argument.
    """

    def __init__(self, gtab):
        ReconstModel.__init__(self, gtab)

    def fit(self, data):
        """To be implemented by specific odf models"""
        raise NotImplementedError("To be implemented in sub classes")


class OdfFit(ReconstFit):
    def odf(self, sphere):
        """To be implemented but specific odf models"""
        raise NotImplementedError("To be implemented in sub classes")


def gfa(samples):
    r"""The general fractional anisotropy of a function evaluated
    on the unit sphere

    Parameters
    ----------
    samples : ndarray
        Values of data on the unit sphere.

    Returns
    -------
    gfa : ndarray
        GFA evaluated in each entry of the array, along the last dimension.
        An `np.nan` is returned for coordinates that contain all-zeros in
        `samples`.

    Notes
    -----
    The GFA is defined as :footcite:p:`CohenAdad2011`:

    .. math::

        \sqrt{\frac{n \sum_i{(\Psi_i - <\Psi>)^2}}{(n-1) \sum{\Psi_i ^ 2}}}

    Where $\Psi$ is an orientation distribution function sampled discretely on
    the unit sphere and angle brackets denote average over the samples on the
    sphere.

    References
    ----------
    .. footbibliography::
    """
    diff = samples - samples.mean(-1)[..., None]
    n = samples.shape[-1]
    numer = np.array([n * (diff**2).sum(-1)])
    denom = np.array([(n - 1) * (samples**2).sum(-1)])
    result = np.ones_like(denom) * np.nan
    idx = np.where(denom > 0)
    result[idx] = np.sqrt(numer[idx] / denom[idx])
    return result.squeeze()


@warning_for_keywords()
def minmax_normalize(samples, *, out=None):
    """Min-max normalization of a function evaluated on the unit sphere

    Normalizes samples to ``(samples - min(samples)) / (max(samples) -
    min(samples))`` for each unit sphere.

    Parameters
    ----------
    samples : ndarray (..., N)
        N samples on a unit sphere for each point, stored along the last axis
        of the array.
    out : ndrray (..., N), optional
        An array to store the normalized samples.

    Returns
    -------
    out : ndarray, (..., N)
        Normalized samples.

    """
    if out is None:
        dtype = np.common_type(np.empty(0, "float32"), samples)
        out = np.array(samples, dtype=dtype, copy=True)
    else:
        out[:] = samples

    sample_mins = np.min(samples, -1)[..., None]
    sample_maxes = np.max(samples, -1)[..., None]
    out -= sample_mins
    out /= sample_maxes - sample_mins
    return out



import warnings

import numpy as np
import numpy.testing as npt

from dipy.core.sphere import unit_octahedron
from dipy.data import default_sphere
from dipy.direction.pmf import SHCoeffPmfGen, SimplePmfGen
from dipy.reconst.shm import (
    SphHarmFit,
    SphHarmModel,
    descoteaux07_legacy_msg,
)
from dipy.tracking.propspeed import ndarray_offset, eudx_both_directions
from dipy.tracking.propspeed cimport (
    deterministic_propagator,
    probabilistic_propagator,
    parallel_transport_propagator,
)

from dipy.tracking.tracker_parameters import generate_tracking_parameters, FAIL
from dipy.tracking.tests.test_tractogen import get_fast_tracking_performances
from dipy.utils.fast_numpy cimport RNGState, seed_rng


def test_tracker_deterministic():
    # Test the probabilistic tracker function
    cdef double[:] stream_data = np.zeros(3, dtype=float)
    cdef double[:] point
    cdef double[:] direction
    cdef RNGState rng

    seed_rng(&rng, 12345)
    class SillyModel(SphHarmModel):
        sh_order_max = 4

        def fit(self, data, mask=None):
            coeff = np.zeros(data.shape[:-1] + (15,))
            return SphHarmFit(self, coeff, mask=None)

    model = SillyModel(gtab=None)
    data = np.zeros((3, 3, 3, 7))
    sphere = unit_octahedron

    params = generate_tracking_parameters("det",
                                          max_len=500,
                                          step_size=0.2,
                                          voxel_size=np.ones(3),
                                          max_angle=20)

    # Test if the tracking works on different dtype of the same data.
    for dtype in [np.float32, np.float64]:
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )
            fit = model.fit(data.astype(dtype))
            sh_pmf_gen = SHCoeffPmfGen(fit.shm_coeff, sphere, 'descoteaux07')
            sf_pmf_gen = SimplePmfGen(fit.odf(sphere), sphere)

        point = np.zeros(3)
        direction = unit_octahedron.vertices[0].copy()

        # Test using SH pmf
        status = deterministic_propagator(&point[0],
                                          &direction[0],
                                          params,
                                          &stream_data[0],
                                          sh_pmf_gen,
                                          &rng)
        npt.assert_equal(status, FAIL)

        # Test using SF pmf
        status = deterministic_propagator(&point[0],
                                          &direction[0],
                                          params,
                                          &stream_data[0],
                                          sf_pmf_gen,
                                          &rng)
        npt.assert_equal(status, FAIL)


def test_deterministic_performances():
    # Test deterministic tracker on the DiSCo dataset
    params = generate_tracking_parameters("det",
                                          max_len=500,
                                          step_size=0.2,
                                          voxel_size=np.ones(3),
                                          max_angle=20)
    r = get_fast_tracking_performances(params, nbr_seeds=5000)
    npt.assert_(r > 0.85, msg="Deterministic tracker has a low performance "
                              "score: " + str(r))


def test_tracker_probabilistic():
    # Test the probabilistic tracker function
    cdef double[:] stream_data = np.zeros(3, dtype=float)
    cdef double[:] point
    cdef double[:] direction
    cdef RNGState rng

    seed_rng(&rng, 12345)

    class SillyModel(SphHarmModel):
        sh_order_max = 4

        def fit(self, data, mask=None):
            coeff = np.zeros(data.shape[:-1] + (15,))
            return SphHarmFit(self, coeff, mask=None)

    model = SillyModel(gtab=None)
    data = np.zeros((3, 3, 3, 7))
    sphere = unit_octahedron

    params = generate_tracking_parameters("prob",
                                          max_len=500,
                                          step_size=0.2,
                                          voxel_size=np.ones(3),
                                          max_angle=20)

    # Test if the tracking works on different dtype of the same data.
    for dtype in [np.float32, np.float64]:
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )
            fit = model.fit(data.astype(dtype))
            sh_pmf_gen = SHCoeffPmfGen(fit.shm_coeff, sphere, 'descoteaux07')
            sf_pmf_gen = SimplePmfGen(fit.odf(sphere), sphere)

        point = np.zeros(3)
        direction = unit_octahedron.vertices[0].copy()

        # Test using SH pmf
        status = probabilistic_propagator(&point[0],
                                          &direction[0],
                                          params,
                                          &stream_data[0],
                                          sh_pmf_gen,
                                          &rng)
        npt.assert_equal(status, FAIL)

        # Test using SF pmf
        status = probabilistic_propagator(&point[0],
                                          &direction[0],
                                          params,
                                          &stream_data[0],
                                          sf_pmf_gen,
                                          &rng)
        npt.assert_equal(status, FAIL)


def test_probabilistic_performances():
    # Test probabilistic tracker on the DiSCo dataset
    params = generate_tracking_parameters("prob",
                                          max_len=500,
                                          step_size=0.2,
                                          voxel_size=np.ones(3),
                                          max_angle=20)
    r = get_fast_tracking_performances(params, nbr_seeds=10000)
    npt.assert_(r > 0.85, msg="Probabilistic tracker has a low performance "
                              "score: " + str(r))


def test_tracker_ptt():
    # Test the probabilistic tracker function
    cdef double[:] stream_data = np.zeros(100, dtype=float)
    cdef double[:] point
    cdef double[:] direction
    cdef RNGState rng

    seed_rng(&rng, 12345)

    class SillyModel(SphHarmModel):
        sh_order_max = 4

        def fit(self, data, mask=None):
            coeff = np.zeros(data.shape[:-1] + (15,))
            return SphHarmFit(self, coeff, mask=None)

    model = SillyModel(gtab=None)
    data = np.zeros((3, 3, 3, 7))
    sphere = unit_octahedron

    params = generate_tracking_parameters("ptt",
                                          max_len=500,
                                          step_size=0.2,
                                          voxel_size=np.ones(3),
                                          max_angle=20,
                                          probe_quality=3)

    # Test if the tracking works on different dtype of the same data.
    for dtype in [np.float32, np.float64]:
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )
            fit = model.fit(data.astype(dtype))
            sh_pmf_gen = SHCoeffPmfGen(fit.shm_coeff, sphere, 'descoteaux07')
            sf_pmf_gen = SimplePmfGen(fit.odf(sphere), sphere)

        point = np.zeros(3)
        direction = unit_octahedron.vertices[0].copy()

        # Test using SH pmf
        status = parallel_transport_propagator(&point[0],
                                               &direction[0],
                                               params,
                                               &stream_data[0],
                                               sh_pmf_gen,
                                               &rng)
        npt.assert_equal(status, FAIL)

        # Test using SF pmf
        status = parallel_transport_propagator(&point[0],
                                               &direction[0],
                                               params,
                                               &stream_data[0],
                                               sf_pmf_gen,
                                               &rng)
        npt.assert_equal(status, FAIL)


def test_ptt_performances():
    # Test ptt tracker on the DiSCo dataset

    params = generate_tracking_parameters("ptt",
                                          max_len=500,
                                          step_size=0.2,
                                          voxel_size=np.ones(3),
                                          max_angle=15,
                                          probe_quality=4)

    r = get_fast_tracking_performances(params, nbr_seeds=5000)
    npt.assert_(r > 0.85, msg="PTT tracker has a low performance "
                              "score: " + str(r))


def stepped_1d(arr_1d):
    # Make a version of `arr_1d` which is not contiguous
    return np.vstack((arr_1d, arr_1d)).ravel(order="F")[::2]


def test_offset():
    # Test ndarray_offset function
    for dt in (np.int32, np.float64):
        index = np.array([1, 1], dtype=np.intp)
        A = np.array([[1, 0, 0], [0, 2, 0], [0, 0, 3]], dtype=dt)
        strides = np.array(A.strides, np.intp)
        i_size = A.dtype.itemsize
        npt.assert_equal(ndarray_offset(index, strides, 2, i_size), 4)
        npt.assert_equal(A.ravel()[4], A[1, 1])
        # Index and strides arrays must be C-continuous. Test this is enforced
        # by using non-contiguous versions of the input arrays.
        npt.assert_raises(ValueError, ndarray_offset, stepped_1d(index), strides, 2, i_size)
        npt.assert_raises(ValueError, ndarray_offset, index, stepped_1d(strides), 2, i_size)


def test_eudx_both_directions_errors():
    # Test error conditions for both directions function
    sphere = default_sphere
    seed = np.zeros(3, np.float64)
    qa = np.zeros((4, 5, 6, 7), np.float64)
    ind = qa.copy()
    # All of seed, qa, ind, odf_vertices must be C-contiguous.  Check by
    # passing in versions that aren't C contiguous
    npt.assert_raises(
        ValueError,
        eudx_both_directions,
        stepped_1d(seed),
        0,
        qa,
        ind,
        sphere.vertices,
        0.5,
        0.1,
        1.0,
        1.0,
        2,
    )
    npt.assert_raises(
        ValueError,
        eudx_both_directions,
        seed,
        0,
        qa[..., ::2],
        ind,
        sphere.vertices,
        0.5,
        0.1,
        1.0,
        1.0,
        2,
    )
    npt.assert_raises(
        ValueError,
        eudx_both_directions,
        seed,
        0,
        qa,
        ind[..., ::2],
        sphere.vertices,
        0.5,
        0.1,
        1.0,
        1.0,
        2,
    )
    npt.assert_raises(
        ValueError,
        eudx_both_directions,
        seed,
        0,
        qa,
        ind,
        sphere.vertices[::2],
        0.5,
        0.1,
        1.0,
        1.0,
        2,
    )



"""
==================================
Advanced interactive visualization
==================================

In DIPY_ we created a thin interface to access many of the capabilities
available in the FURY 3D visualization library :footcite:p:`Garyfallidis2021`
but tailored to the needs of structural and diffusion imaging.

Let's start by importing the necessary modules.
"""

import numpy as np

from dipy.data.fetcher import fetch_bundles_2_subjects, read_bundles_2_subjects
from dipy.tracking.streamline import Streamlines
from dipy.viz import actor, ui, window

###############################################################################
# In ``window`` we have all the objects that connect what needs to be rendered
# to the display or the disk e.g., for saving screenshots. So, there you will
# find key objects and functions like the ``Scene`` class which holds and
# provides access to all the actors and the ``show`` function which displays
# what is in the scene on a window. Also, this module provides access to
# functions for opening/saving dialogs and printing screenshots
# (see ``snapshot``).
#
# In the ``actor`` module we can find all the different primitives e.g.,
# streamtubes, lines, image slices, etc.
#
# In the ``ui`` module we have some other objects which allow to add buttons
# and sliders and these interact both with windows and actors. Because of
# this they need input from the operating system so they can process events.
#
# Let's get started. In this tutorial, we will visualize some bundles
# together with FA or T1. We will be able to change the slices using
# a ``LineSlider2D`` widget.
#
# First we need to fetch and load some datasets.

fetch_bundles_2_subjects()

###############################################################################
# The following function outputs a dictionary with the required bundles e.g.
# ``af left`` (left arcuate fasciculus) and maps, e.g. FA for a specific
# subject.

res = read_bundles_2_subjects(
    subj_id="subj_1", metrics=["t1", "fa"], bundles=["af.left", "cst.right", "cc_1"]
)

###############################################################################
# We will use 3 bundles, FA and the affine transformation that brings the voxel
# coordinates to world coordinates (RAS 1mm).

streamlines = Streamlines(res["af.left"])
streamlines.extend(res["cst.right"])
streamlines.extend(res["cc_1"])

data = res["fa"]
shape = data.shape
affine = res["affine"]

###############################################################################
# With our current design it is easy to decide in which space you want the
# streamlines and slices to appear. The default we have here is to appear in
# world coordinates (RAS 1mm).

world_coords = True

###############################################################################
# If we want to see the objects in native space we need to make sure that all
# objects which are currently in world coordinates are transformed back to
# native space using the inverse of the affine.

if not world_coords:
    from dipy.tracking.streamline import transform_streamlines

    streamlines = transform_streamlines(streamlines, np.linalg.inv(affine))

###############################################################################
# Now we create, a ``Scene`` object and add the streamlines using the ``line``
# function and an image plane using the ``slice`` function.

scene = window.Scene()
stream_actor = actor.line(streamlines)

if not world_coords:
    image_actor_z = actor.slicer(data, affine=np.eye(4))
else:
    image_actor_z = actor.slicer(data, affine=affine)

###############################################################################
# We can also change also the opacity of the slicer.

slicer_opacity = 0.6
image_actor_z.opacity(slicer_opacity)

###############################################################################
# We can add additional slicers by copying the original and adjusting the
# ``display_extent``.

image_actor_x = image_actor_z.copy()
x_midpoint = int(np.round(shape[0] / 2))
image_actor_x.display_extent(x_midpoint, x_midpoint, 0, shape[1] - 1, 0, shape[2] - 1)

image_actor_y = image_actor_z.copy()
y_midpoint = int(np.round(shape[1] / 2))
image_actor_y.display_extent(0, shape[0] - 1, y_midpoint, y_midpoint, 0, shape[2] - 1)

###############################################################################
# Connect the actors with the Scene.

scene.add(stream_actor)
scene.add(image_actor_z)
scene.add(image_actor_x)
scene.add(image_actor_y)

###############################################################################
# Now we would like to change the position of each ``image_actor`` using a
# slider. The sliders are widgets which require access to different areas of
# the visualization pipeline and therefore we don't recommend using them with
# ``show``. The more appropriate way is to use them with the ``ShowManager``
# object which allows accessing the pipeline in different areas. Here is how:

show_m = window.ShowManager(scene=scene, size=(1200, 900))
show_m.initialize()

###############################################################################
# After we have initialized the ``ShowManager`` we can go ahead and create
# sliders to move the slices and change their opacity.

line_slider_z = ui.LineSlider2D(
    min_value=0,
    max_value=shape[2] - 1,
    initial_value=shape[2] / 2,
    text_template="{value:.0f}",
    length=140,
)

line_slider_x = ui.LineSlider2D(
    min_value=0,
    max_value=shape[0] - 1,
    initial_value=shape[0] / 2,
    text_template="{value:.0f}",
    length=140,
)

line_slider_y = ui.LineSlider2D(
    min_value=0,
    max_value=shape[1] - 1,
    initial_value=shape[1] / 2,
    text_template="{value:.0f}",
    length=140,
)

opacity_slider = ui.LineSlider2D(
    min_value=0.0, max_value=1.0, initial_value=slicer_opacity, length=140
)

###############################################################################
# Now we will write callbacks for the sliders and register them.


def change_slice_z(slider):
    z = int(np.round(slider.value))
    image_actor_z.display_extent(0, shape[0] - 1, 0, shape[1] - 1, z, z)


def change_slice_x(slider):
    x = int(np.round(slider.value))
    image_actor_x.display_extent(x, x, 0, shape[1] - 1, 0, shape[2] - 1)


def change_slice_y(slider):
    y = int(np.round(slider.value))
    image_actor_y.display_extent(0, shape[0] - 1, y, y, 0, shape[2] - 1)


def change_opacity(slider):
    slicer_opacity = slider.value
    image_actor_z.opacity(slicer_opacity)
    image_actor_x.opacity(slicer_opacity)
    image_actor_y.opacity(slicer_opacity)


line_slider_z.on_change = change_slice_z
line_slider_x.on_change = change_slice_x
line_slider_y.on_change = change_slice_y
opacity_slider.on_change = change_opacity

###############################################################################
# We'll also create text labels to identify the sliders.


def build_label(text):
    label = ui.TextBlock2D()
    label.message = text
    label.font_size = 18
    label.font_family = "Arial"
    label.justification = "left"
    label.bold = False
    label.italic = False
    label.shadow = False
    label.background_color = (0, 0, 0)
    label.color = (1, 1, 1)

    return label


line_slider_label_z = build_label(text="Z Slice")
line_slider_label_x = build_label(text="X Slice")
line_slider_label_y = build_label(text="Y Slice")
opacity_slider_label = build_label(text="Opacity")

###############################################################################
# Now we will create a ``panel`` to contain the sliders and labels.

panel = ui.Panel2D(size=(300, 200), color=(1, 1, 1), opacity=0.1, align="right")
panel.center = (1030, 120)

panel.add_element(line_slider_label_x, (0.1, 0.75))
panel.add_element(line_slider_x, (0.38, 0.75))
panel.add_element(line_slider_label_y, (0.1, 0.55))
panel.add_element(line_slider_y, (0.38, 0.55))
panel.add_element(line_slider_label_z, (0.1, 0.35))
panel.add_element(line_slider_z, (0.38, 0.35))
panel.add_element(opacity_slider_label, (0.1, 0.15))
panel.add_element(opacity_slider, (0.38, 0.15))

scene.add(panel)

###############################################################################
# Then, we can render all the widgets and everything else in the screen and
# start the interaction using ``show_m.start()``.
#
#
# However, if you change the window size, the panel will not update its
# position properly. The solution to this issue is to update the position of
# the panel using its ``re_align`` method every time the window size changes.

global size
size = scene.GetSize()


def win_callback(obj, event):
    global size
    if size != obj.GetSize():
        size_old = size
        size = obj.GetSize()
        size_change = [size[0] - size_old[0], 0]
        panel.re_align(size_change)


show_m.initialize()

###############################################################################
# Finally, please set the following variable to ``True`` to interact with the
# datasets in 3D.

interactive = False

scene.zoom(1.5)
scene.reset_clipping_range()

if interactive:
    show_m.add_window_callback(win_callback)
    show_m.render()
    show_m.start()

else:
    window.record(
        scene=scene,
        out_path="bundles_and_3_slices.png",
        size=(1200, 900),
        reset_camera=False,
    )

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# A few bundles with interactive slicing.

del show_m

###############################################################################
# References
# ----------
#
# .. footbibliography::
#



# dipy documentation build configuration file, created by
# sphinx-quickstart on Thu Feb  4 15:23:20 2010.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import datetime
import os
import re
import sys
import time

# Doc generation depends on being able to import dipy
try:
    import dipy
except ImportError as e:
    raise RuntimeError("Cannot import dipy, please investigate") from e

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.append(os.path.abspath("sphinxext"))

# -- General configuration -----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.todo",
    "sphinx.ext.ifconfig",
    "math_dollar",  # has to go before numpydoc
    "numpydoc",
    "prepare_gallery",
    "sphinx_gallery.gen_gallery",
    "sphinxcontrib.bibtex",
    "github",
    "sphinx_design",
]

numpydoc_show_class_members = True
numpydoc_class_members_toctree = False
numpydoc_show_inherited_class_members = {
    "dipy.viz.horizon.visualizer.peak.PeakActor": False
}

autodoc_skip_members = [
    "docstring_addendum",
]

# Sphinx extension for BibTeX style citations.
# https://github.com/mcmtroffaes/sphinxcontrib-bibtex
bibtex_bibfiles = ["references.bib"]
# bibtex_default_style = 'plain'

# ghissue config
github_project_url = "https://github.com/dipy/dipy"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8'

# The master toctree document.
master_doc = "index"

build_date = datetime.datetime.fromtimestamp(
    int(os.environ.get("SOURCE_DATE_EPOCH", time.time())),
    tz=datetime.timezone.utc,
)

# General information about the project.
project = "dipy"
copyright = (
    f"Copyright 2008-{build_date.year},DIPY developers."
    f" Created using Grg Sphinx Theme and PyData Sphinx Theme."
)

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = dipy.__version__
# The full version, including alpha/beta/rc tags.
release = version

# Include common links
# We don't use this any more because it causes conflicts with the gitwash docs
# rst_epilog = open('links_names.inc', 'rt').read()

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of documents that shouldn't be included in the build.
# unused_docs = []

# List of directories, relative to source directory, that shouldn't be searched
# for source files.
exclude_patterns = [
    "_build",
    "examples",
    "examples_revamped",
    "reconstruction_models_list.rst",
    "tractography_methods_list.rst",
]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme = "grg_sphinx_theme"

# The style sheet to use for HTML and HTML Help pages. A file of that name
# must exist either in Sphinx' static/ path, or in one of the custom paths
# given in html_static_path.
html_style = "css/dipy.css"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
doc_version = "development" if "dev" in version else version
html_theme_options = {
    "switcher": {
        "json_url": "https://docs.dipy.org/dev/_static/version_switcher.json",
        "version_match": doc_version,
    },
    "check_switcher": False,
    "show_version_warning_banner": True,
    "navbar_end": ["search-field.html", "version-switcher", "navbar-icon-links.html"],
    "secondary_sidebar_items": ["page-toc"],
    "show_toc_level": 1,
    "navbar_center": ["components/navbar-links.html"],
    "navbar_links": [
        {
            "name": "Docs",
            "children": [
                {
                    "name": "Overview",
                    "url": "index",
                },
                {
                    "name": "Tutorials",
                    "url": "examples_built/index",
                },
                {
                    "name": "Recipes",
                    "url": "recipes",
                },
                {
                    "name": "CLI / Workflows",
                    "url": "interfaces/index",
                },
                {
                    "name": "API",
                    "url": "reference/index",
                },
                {
                    "name": "CLI API",
                    "url": "reference_cmd/index",
                },
            ],
        },
        {
            "name": "Workshops",
            "sections": [
                {
                    "name": "Latest",
                    "children": [
                        {
                            "name": "DIPY Workshop 2024",
                            "url": "https://dipy.org/workshops/dipy-workshop-2024",
                            "link_type": "external",
                        }
                    ],
                },
                {
                    "name": "Past",
                    "children": [
                        {
                            "name": "DIPY Workshop 2023",
                            "url": "https://dipy.org/workshops/dipy-workshop-2023",
                            "link_type": "external",
                        },
                        {
                            "name": "DIPY Workshop 2022",
                            "url": "https://dipy.org/workshops/dipy-workshop-2022",
                            "link_type": "external",
                        },
                        {
                            "name": "DIPY Workshop 2021",
                            "url": "https://dipy.org/workshops/dipy-workshop-2021",
                            "link_type": "external",
                        },
                        {
                            "name": "DIPY Workshop 2020",
                            "url": "https://dipy.org/workshops/dipy-workshop-2020",
                            "link_type": "external",
                        },
                        {
                            "name": "DIPY Workshop 2019",
                            "url": "https://dipy.org/workshops/dipy-workshop-2019",
                            "link_type": "external",
                        },
                    ],
                },
            ],
        },
        {
            "name": "Community",
            "sections": [
                {
                    "name": "News",
                    "children": [
                        {
                            "name": "Calendar",
                            "url": "https://dipy.org/calendar",
                            "link_type": "inter",
                        },
                        {
                            "name": "Newsletters",
                            "url": "https://mail.python.org/mailman3/lists/dipy.python.org/",
                            "link_type": "external",
                        },
                        {
                            "name": "Blog",
                            "url": "https://dipy.org/blog",
                            "link_type": "inter",
                        },
                        {
                            "name": "Youtube",
                            "url": "https://www.youtube.com/c/diffusionimaginginpython",
                            "link_type": "external",
                        },
                    ],
                },
                {
                    "name": "Help",
                    "children": [
                        {
                            "name": "Live Chat (Gitter)",
                            "url": "https://app.gitter.im/#/room/%23dipy_dipy:gitter.im",
                            "link_type": "external",
                        },
                        {
                            "name": "Github Discussions",
                            "url": "https://github.com/dipy/dipy/discussions",
                            "link_type": "external",
                        },
                    ],
                },
            ],
        },
        {
            "name": "About",
            "children": [
                {"name": "Team", "url": "https://dipy.org/team", "link_type": "inter"},
                {
                    "name": "FAQ",
                    "url": "faq",
                },
                {
                    "name": "Mission Statement",
                    "url": "user_guide/mission",
                },
                {
                    "name": "Releases",
                    "url": "stateoftheart",
                },
                {
                    "name": "Cite",
                    "url": "cite",
                },
                {
                    "name": "Glossary",
                    "url": "glossary",
                },
            ],
        },
    ],
    # To remove search icon
    "navbar_persistent": "",
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/dipy",
            "icon": "fa-brands fa-github",
        },
        {
            "name": "Twitter/X",
            "url": "https://twitter.com/dipymri",
            "icon": "fa-brands fa-twitter",
        },
        {
            "name": "YouTube",
            "url": "https://www.youtube.com/c/diffusionimaginginpython",
            "icon": "fa-brands fa-youtube",
        },
        {
            "name": "LinkedIn",
            "url": "https://www.linkedin.com/company/dipy/",
            "icon": "fa-brands fa-linkedin",
        },
    ],
    "logo": {
        "image_dark": "_static/images/logos/dipy-logo.png",
        "alt_text": "DIPY",
        "link": "https://dipy.org",
    },
    "footer_start": ["components/footer-sign-up.html"],
    "footer_signup_data": {
        "heading": "Never miss an update from us!",
        "sub_heading": "Don't worry! we are not going to spam you.",
    },
    "footer_end": ["components/footer-sections.html"],
    "footer_links": [
        {
            "title": "About",
            "links": [
                {
                    "name": "Developers",
                    "link": "https://dipy.org/team",
                    "link_type": "inter",
                },
                {
                    "name": "Support",
                    "link": "https://github.com/dipy/dipy/discussions",
                    "link_type": "external",
                },
                {"name": "Download", "link": "user_guide/installation"},
                {"name": "Get Started", "link": "user_guide/getting_started"},
                {"name": "Tutorials", "link": "examples_built/index"},
                {
                    "name": "Videos",
                    "link": "https://www.youtube.com/c/diffusionimaginginpython",
                    "link_type": "external",
                },
            ],
        },
        {
            "title": "Friends",
            "links": [
                {
                    "name": "Nipy Projects",
                    "link": "http://nipy.org/",
                    "link_type": "external",
                },
                {"name": "FURY", "link": "http://fury.gl/", "link_type": "external"},
                {
                    "name": "Nibabel",
                    "link": "http://nipy.org/nibabel",
                    "link_type": "external",
                },
                {
                    "name": "Tortoise",
                    "link": "https://tortoise.nibib.nih.gov/",
                    "link_type": "external",
                },
            ],
        },
        {
            "title": "Support",
            "links": [
                {
                    "name": "The department of Intelligent Systems Engineering of Indiana University",  # noqa: E501
                    "link": "https://engineering.indiana.edu/",
                    "link_type": "external",
                },
                {
                    "name": "The National Institute of Biomedical Imaging and Bioengineering, NIH",  # noqa: E501
                    "link": "https://www.nibib.nih.gov/",
                    "link_type": "external",
                },
                {
                    "name": "The Gordon and Betty Moore Foundation and the Alfred P. Sloan Foundation, through the University of Washington eScience Institute Data Science Environment",  # noqa: E501
                    "link": "https://escience.washington.edu",
                    "link_type": "external",
                },
                {
                    "name": "Google supported DIPY through the Google Summer of Code Program (2015-2024)",  # noqa: E501
                    "link": "https://summerofcode.withgoogle.com/",
                    "link_type": "external",
                },
            ],
        },
    ],
    "footer_copyright": copyright,
}

html_theme_options["analytics"] = {
    "google_analytics_id": "G-D610GKJZRC",
}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/images/logos/dipy-logo.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/images/logos/dipy-favicon.png"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {"index": []}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# Setting to false fixes double module listing under header
html_use_modindex = False

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = ''

# Output file base name for HTML help builder.
htmlhelp_basename = "dipydoc"


# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
# latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    (
        "index",
        "dipy.tex",
        "dipy Documentation",
        "Eleftherios Garyfallidis, Ian Nimmo-Smith, Matthew Brett",
        "manual",
    ),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# Additional stuff for the LaTeX preamble.
latex_preamble = r"""
\usepackage{amsfonts}
"""

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_use_modindex = True


# -- Options for sphinx gallery -------------------------------------------
from docimage_scrap import ImageFileScraper
from prepare_gallery import folder_explicit_order
from sphinx_gallery.sorting import ExplicitOrder

sc = ImageFileScraper()
ordered_folders = [f"examples_revamped/{f}" for f in folder_explicit_order()]

sphinx_gallery_conf = {
    "doc_module": ("dipy",),
    # path to your examples scripts
    "examples_dirs": [
        "examples_revamped",
    ],
    # path where to save gallery generated examples
    "gallery_dirs": [
        "examples_built",
    ],
    "subsection_order": ExplicitOrder(ordered_folders),
    "image_scrapers": (sc),
    "backreferences_dir": "examples_built",
    "reference_url": {
        "dipy": None,
    },
    "abort_on_example_error": False,
    "filename_pattern": re.escape(os.sep),
    "default_thumb_file": "_static/images/logos/dipy_full_logo.png",
    #  'pypandoc': {'extra_args': ['--mathjax',]},
}

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {"python": ("https://docs.python.org/3/", None)}


def setup(app):
    def skip_doc_element(app, what, name, obj, would_skip, options):
        if name in autodoc_skip_members:
            return True
        return would_skip

    app.connect("autodoc-skip-member", skip_doc_element)



"""
=================================
Parallel reconstruction using CSD
=================================

This example shows how to use parallelism (multiprocessing) using
``peaks_from_model`` in order to speedup the signal reconstruction
process. For this example will we use the same initial steps
as we used in :ref:`sphx_glr_examples_built_reconstruction_reconst_csd.py`.

Import modules, fetch and read data, apply the mask and calculate the response
function.
"""

import time

from dipy.core.gradients import gradient_table
from dipy.data import default_sphere, get_fnames
from dipy.direction import peaks_from_model
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.reconst.csdeconv import ConstrainedSphericalDeconvModel, auto_response_ssst
from dipy.segment.mask import median_otsu

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")

data, affine = load_nifti(hardi_fname)

bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

maskdata, mask = median_otsu(
    data, vol_idx=range(10, 50), median_radius=3, numpass=1, autocrop=False, dilate=2
)

response, ratio = auto_response_ssst(gtab, maskdata, roi_radii=10, fa_thr=0.7)

data = maskdata[:, :, 33:37]
mask = mask[:, :, 33:37]

###############################################################################
# Now we are ready to import the CSD model and fit the datasets.

csd_model = ConstrainedSphericalDeconvModel(gtab, response)

###############################################################################
# Compute the CSD-based ODFs using ``peaks_from_model``. This function has a
# parameter called ``parallel`` which allows for the voxels to be processed in
# parallel. If ``num_processes`` is None it will figure out automatically the
# number of CPUs available in your system. Alternatively, you can set
# ``num_processes`` manually. Here, we show an example where we compare the
# duration of execution with or without parallelism.

start_time = time.time()
csd_peaks_parallel = peaks_from_model(
    model=csd_model,
    data=data,
    sphere=default_sphere,
    relative_peak_threshold=0.5,
    min_separation_angle=25,
    mask=mask,
    return_sh=True,
    return_odf=False,
    normalize_peaks=True,
    npeaks=5,
    parallel=True,
    num_processes=2,
)

time_parallel = time.time() - start_time
print(f"peaks_from_model using 2 processes ran in : {time_parallel} seconds")

start_time = time.time()
csd_peaks = peaks_from_model(
    model=csd_model,
    data=data,
    sphere=default_sphere,
    relative_peak_threshold=0.5,
    min_separation_angle=25,
    mask=mask,
    return_sh=True,
    return_odf=False,
    normalize_peaks=True,
    npeaks=5,
    parallel=False,
    num_processes=None,
)

time_single = time.time() - start_time
print(f"peaks_from_model ran in : {time_single} seconds")

print(f"Speedup factor : {time_single / time_parallel}")

###############################################################################
# In Windows if you get a runtime error about frozen executable please start
# your script by adding your code above in a ``main`` function and use::
#
#    if __name__ == '__main__':
#        import multiprocessing
#        multiprocessing.freeze_support()
#        main()



"""Function for recording and reporting deprecations.

Note
-----
this file is copied (with minor modifications) from the Nibabel.
https://github.com/nipy/nibabel. See COPYING file distributed along with
the Nibabel package for the copyright and license terms.

"""

import functools
from inspect import signature
import re
import warnings

from packaging.version import parse as version_cmp

from dipy import __version__
from dipy.testing.decorators import warning_for_keywords

_LEADING_WHITE = re.compile(r"^(\s*)")


class ExpiredDeprecationError(RuntimeError):
    """Error for expired deprecation.

    Error raised when a called function or method has passed out of its
    deprecation period.

    """

    pass


class ArgsDeprecationWarning(DeprecationWarning):
    """Warning for args deprecation.

    Warning raised when a function or method argument has changed or removed.

    """

    pass


def _ensure_cr(text):
    """Remove trailing whitespace and add carriage return.

    Ensures that `text` always ends with a carriage return
    """
    return text.rstrip() + "\n"


def _add_dep_doc(old_doc, dep_doc):
    """Add deprecation message `dep_doc` to docstring in `old_doc`.

    Parameters
    ----------
    old_doc : str
        Docstring from some object.
    dep_doc : str
        Deprecation warning to add to top of docstring, after initial line.

    Returns
    -------
    new_doc : str
        `old_doc` with `dep_doc` inserted after any first lines of docstring.

    """
    dep_doc = _ensure_cr(dep_doc)
    if not old_doc:
        return dep_doc
    old_doc = _ensure_cr(old_doc)
    old_lines = old_doc.splitlines()
    new_lines = []
    for _line_no, line in enumerate(old_lines):
        if line.strip():
            new_lines.append(line)
        else:
            break
    next_line = _line_no + 1
    if next_line >= len(old_lines):
        # nothing following first paragraph, just append message
        return f"{old_doc}\n{dep_doc}"
    indent = _LEADING_WHITE.match(old_lines[next_line]).group()
    dep_lines = [indent + L for L in [""] + dep_doc.splitlines() + [""]]
    return "\n".join(new_lines + dep_lines + old_lines[next_line:]) + "\n"


@warning_for_keywords()
def cmp_pkg_version(version_str, *, pkg_version_str=__version__):
    """Compare `version_str` to current package version.

    Parameters
    ----------
    version_str : str
        Version string to compare to current package version
    pkg_version_str : str, optional
        Version of our package.  Optional, set from ``__version__`` by default.

    Returns
    -------
    version_cmp : int
        1 if `version_str` is a later version than `pkg_version_str`, 0 if
        same, -1 if earlier.

    Examples
    --------
    >>> cmp_pkg_version('1.2.1', pkg_version_str='1.2.0')
    1
    >>> cmp_pkg_version('1.2.0dev', pkg_version_str='1.2.0')
    -1

    """
    if any(re.match(r"^[a-z, A-Z]", v) for v in [version_str, pkg_version_str]):
        msg = f"Invalid version {version_str} or {pkg_version_str}"
        raise ValueError(msg)
    elif version_cmp(version_str) > version_cmp(pkg_version_str):
        return 1
    elif version_cmp(version_str) == version_cmp(pkg_version_str):
        return 0
    else:
        return -1


@warning_for_keywords()
def is_bad_version(version_str, *, version_comparator=cmp_pkg_version):
    """Return True if `version_str` is too high."""
    return version_comparator(version_str) == -1


@warning_for_keywords()
def deprecate_with_version(
    message,
    *,
    since="",
    until="",
    version_comparator=cmp_pkg_version,
    warn_class=DeprecationWarning,
    error_class=ExpiredDeprecationError,
):
    """Return decorator function function for deprecation warning / error.

    The decorated function / method will:

    * Raise the given `warning_class` warning when the function / method gets
      called, up to (and including) version `until` (if specified);
    * Raise the given `error_class` error when the function / method gets
      called, when the package version is greater than version `until` (if
      specified).

    Parameters
    ----------
    message : str
        Message explaining deprecation, giving possible alternatives.
    since : str, optional
        Released version at which object was first deprecated.
    until : str, optional
        Last released version at which this function will still raise a
        deprecation warning.  Versions higher than this will raise an
        error.
    version_comparator : callable
        Callable accepting string as argument, and return 1 if string
        represents a higher version than encoded in the `version_comparator`, 0
        if the version is equal, and -1 if the version is lower.  For example,
        the `version_comparator` may compare the input version string to the
        current package version string.
    warn_class : class, optional
        Class of warning to generate for deprecation.
    error_class : class, optional
        Class of error to generate when `version_comparator` returns 1 for a
        given argument of ``until``.

    Returns
    -------
    deprecator : func
        Function returning a decorator.

    """
    messages = [message]
    if (since, until) != ("", ""):
        messages.append("")
    if since:
        messages.append(f"* deprecated from version: {since}")
    if until:
        raise_will_raise = "Raises" if is_bad_version(until) else "Will raise"
        messages.append(f"* {raise_will_raise} {error_class} as of version: {until}")
    message = "\n".join(messages)

    def deprecator(func):
        @functools.wraps(func)
        def deprecated_func(*args, **kwargs):
            if until and is_bad_version(until, version_comparator=version_comparator):
                raise error_class(message)
            warnings.warn(message, warn_class, stacklevel=2)
            return func(*args, **kwargs)

        deprecated_func.__doc__ = _add_dep_doc(deprecated_func.__doc__, message)
        return deprecated_func

    return deprecator


@warning_for_keywords()
def deprecated_params(
    old_name,
    *,
    new_name=None,
    since="",
    until="",
    version_comparator=cmp_pkg_version,
    arg_in_kwargs=False,
    warn_class=ArgsDeprecationWarning,
    error_class=ExpiredDeprecationError,
    alternative="",
):
    """Deprecate a *renamed* or *removed* function argument.

    The decorator assumes that the argument with the ``old_name`` was removed
    from the function signature and the ``new_name`` replaced it at the
    **same position** in the signature.  If the ``old_name`` argument is
    given when calling the decorated function the decorator will catch it and
    issue a deprecation warning and pass it on as ``new_name`` argument.

    Parameters
    ----------
    old_name : str or list/tuple thereof
        The old name of the argument.
    new_name : str or list/tuple thereof or ``None``, optional
        The new name of the argument. Set this to `None` to remove the
        argument ``old_name`` instead of renaming it.
    since : str or number or list/tuple thereof, optional
        The release at which the old argument became deprecated.
    until : str or number or list/tuple thereof, optional
        Last released version at which this function will still raise a
        deprecation warning.  Versions higher than this will raise an
        error.
    version_comparator : callable
        Callable accepting string as argument, and return 1 if string
        represents a higher version than encoded in the ``version_comparator``,
        0 if the version is equal, and -1 if the version is lower. For example,
        the ``version_comparator`` may compare the input version string to the
        current package version string.
    arg_in_kwargs : bool or list/tuple thereof, optional
        If the argument is not a named argument (for example it
        was meant to be consumed by ``**kwargs``) set this to
        ``True``.  Otherwise the decorator will throw an Exception
        if the ``new_name`` cannot be found in the signature of
        the decorated function.
        Default is ``False``.
    warn_class : warning, optional
        Warning to be issued.
    error_class : Exception, optional
        Error to be issued
    alternative : str, optional
        An alternative function or class name that the user may use in
        place of the deprecated object if ``new_name`` is None. The deprecation
        warning will tell the user about this alternative if provided.

    Raises
    ------
    TypeError
        If the new argument name cannot be found in the function
        signature and arg_in_kwargs was False or if it is used to
        deprecate the name of the ``*args``-, ``**kwargs``-like arguments.
        At runtime such an Error is raised if both the new_name
        and old_name were specified when calling the function and
        "relax=False".

    Notes
    -----
    This function is based on the Astropy (major modification).
    https://github.com/astropy/astropy. See COPYING file distributed along with
    the astropy package for the copyright and license terms.

    Examples
    --------
    The deprecation warnings are not shown in the following examples.
    To deprecate a positional or keyword argument::
    >>> from dipy.utils.deprecator import deprecated_params
    >>> @deprecated_params('sig', new_name='sigma', since='0.3')
    ... def test(sigma):
    ...     return sigma
    >>> test(2)
    2
    >>> test(sigma=2)
    2
    >>> test(sig=2)  # doctest: +SKIP
    2

    It is also possible to replace multiple arguments. The ``old_name``,
    ``new_name`` and ``since`` have to be `tuple` or `list` and contain the
    same number of entries::
    >>> @deprecated_params(['a', 'b'], new_name=['alpha', 'beta'],
    ...                    since=['0.2', 0.4])
    ... def test(alpha, beta):
    ...     return alpha, beta
    >>> test(a=2, b=3)  # doctest: +SKIP
    (2, 3)

    """
    if isinstance(old_name, (list, tuple)):
        # Normalize input parameters
        if not isinstance(arg_in_kwargs, (list, tuple)):
            arg_in_kwargs = [arg_in_kwargs] * len(old_name)
        if not isinstance(since, (list, tuple)):
            since = [since] * len(old_name)
        if not isinstance(until, (list, tuple)):
            until = [until] * len(old_name)
        if not isinstance(new_name, (list, tuple)):
            new_name = [new_name] * len(old_name)

        if (
            len(
                {
                    len(old_name),
                    len(new_name),
                    len(since),
                    len(until),
                    len(arg_in_kwargs),
                }
            )
            != 1
        ):
            raise ValueError("All parameters should have the same length")
    else:
        # To allow a uniform approach later on, wrap all arguments in lists.
        old_name = [old_name]
        new_name = [new_name]
        since = [since]
        until = [until]
        arg_in_kwargs = [arg_in_kwargs]

    def deprecator(function):
        # The named arguments of the function.
        arguments = signature(function).parameters
        positions = [None] * len(old_name)

        for i, (o_name, n_name, in_keywords) in enumerate(
            zip(old_name, new_name, arg_in_kwargs)
        ):
            # Determine the position of the argument.
            if in_keywords:
                continue

            if n_name is not None and n_name not in arguments:
                # In case the argument is not found in the list of arguments
                # the only remaining possibility is that it should be caught
                # by some kind of **kwargs argument.
                msg = f'"{n_name}" was not specified in the function '
                msg += "signature. If it was meant to be part of "
                msg += '"**kwargs" then set "arg_in_kwargs" to "True"'
                raise TypeError(msg)

            key = o_name if n_name is None else n_name
            param = arguments[key]

            if param.kind == param.POSITIONAL_OR_KEYWORD:
                key = o_name if n_name is None else n_name
                positions[i] = list(arguments.keys()).index(key)
            elif param.kind == param.KEYWORD_ONLY:
                # These cannot be specified by position.
                positions[i] = None
            else:
                # positional-only argument, varargs, varkwargs or some
                # unknown type:
                msg = f'cannot replace argument "{n_name}" '
                msg += f"of kind {repr(param.kind)}."
                raise TypeError(msg)

        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            for i, (o_name, n_name) in enumerate(zip(old_name, new_name)):
                messages = [
                    f'"{o_name}" was deprecated',
                ]
                if (since[i], until[i]) != ("", ""):
                    messages.append("")
                if since[i]:
                    messages.append(f"* deprecated from version: {since[i]}")
                if until[i]:
                    raise_will_raise = (
                        "Raises" if is_bad_version(until[i]) else "Will raise"
                    )
                    messages.append(
                        f"* {raise_will_raise} {error_class} as of version: {until[i]}"
                    )
                messages.append("")
                message = "\n".join(messages)

                # The only way to have oldkeyword inside the function is
                # that it is passed as kwarg because the oldkeyword
                # parameter was renamed to newkeyword.
                if o_name in kwargs:
                    value = kwargs.pop(o_name)
                    # Check if the newkeyword was given as well.
                    newarg_in_args = (
                        positions[i] is not None and len(args) > positions[i]
                    )
                    newarg_in_kwargs = n_name in kwargs

                    if newarg_in_args or newarg_in_kwargs:
                        msg = f'cannot specify both "{o_name}"'
                        msg += " (deprecated parameter) and "
                        msg += f'"{n_name}" (new parameter name).'
                        raise TypeError(msg)

                    # Pass the value of the old argument with the
                    # name of the new argument to the function
                    key = n_name or o_name
                    kwargs[key] = value

                    if n_name is not None:
                        message += f'* Use argument "{n_name}" instead.'
                    elif alternative:
                        message += f"* Use {alternative} instead."

                    if until[i] and is_bad_version(
                        until[i], version_comparator=version_comparator
                    ):
                        raise error_class(message)
                    warnings.warn(message, warn_class, stacklevel=2)

                # Deprecated keyword without replacement is given as
                # positional argument.
                elif not n_name and positions[i] and len(args) > positions[i]:
                    if alternative:
                        message += f"* Use {alternative} instead."
                    if until[i] and is_bad_version(
                        until[i], version_comparator=version_comparator
                    ):
                        raise error_class(message)

                    warnings.warn(message, warn_class, stacklevel=2)

            return function(*args, **kwargs)

        return wrapper

    return deprecator



import copy
from warnings import warn

import numpy as np
from scipy.linalg import eigh
from scipy.linalg.lapack import dgesvd as svd

from dipy.denoise.pca_noise_estimate import pca_noise_estimate
from dipy.testing.decorators import warning_for_keywords


def dimensionality_problem_message(arr, num_samples, spr):
    """Message about the number of samples being smaller than one less the
    dimensionality of the data to be denoised.

    Parameters
    ----------
    arr : ndarray
        Data to be denoised.
    num_samples : int
        Number of samples.
    spr : int
        Suggested patch radius.

    Returns
    -------
    str
        Message.
    """

    return (
        f"Number of samples {arr.shape[-1]} - 1 < Dimensionality "
        f"{num_samples}. This might have a performance impact. Increase p"
        f"atch_radius to {spr} to avoid this."
    )


def _pca_classifier(L, nvoxels):
    """Classifies which PCA eigenvalues are related to noise and estimates the
    noise variance

    Parameters
    ----------
    L : array (n,)
        Array containing the PCA eigenvalues in ascending order.
    nvoxels : int
        Number of voxels used to compute L

    Returns
    -------
    var : float
        Estimation of the noise variance
    ncomps : int
        Number of eigenvalues related to noise

    Notes
    -----
    This is based on the algorithm described in :footcite:p:`Veraart2016c`.

    References
    ----------
    .. footbibliography::
    """

    # if num_samples - 1 (to correct for mean subtraction) is less than number
    # of features, discard the zero eigenvalues
    if L.size > nvoxels - 1:
        L = L[-(nvoxels - 1) :]

    # Note that the condition expressed in the while-loop is expressed in terms
    # of the variance of equation (12), not equation (11) as in
    # :footcite:p:`Veraart2016c`. Also, this code implements ascending
    # eigenvalues, unlike :footcite:p:`Veraart2016c`.
    var = np.mean(L)
    c = L.size - 1
    r = L[c] - L[0] - 4 * np.sqrt((c + 1.0) / nvoxels) * var
    while r > 0:
        var = np.mean(L[:c])
        c = c - 1
        r = L[c] - L[0] - 4 * np.sqrt((c + 1.0) / nvoxels) * var
    ncomps = c + 1

    return var, ncomps


def create_patch_radius_arr(arr, pr):
    """Create the patch radius array from the data to be denoised and the patch
    radius.

    Parameters
    ----------
    arr : ndarray
        Data to be denoised.
    pr : int or ndarray
        Patch radius.

    Returns
    -------
    patch_radius : ndarray
        Number of samples.
    """

    patch_radius = copy.deepcopy(pr)

    if isinstance(patch_radius, int):
        patch_radius = np.ones(3, dtype=int) * patch_radius
    if len(patch_radius) != 3:
        raise ValueError("patch_radius should have length 3")
    else:
        patch_radius = np.asarray(patch_radius).astype(int)
    patch_radius[arr.shape[0:3] == np.ones(3)] = 0  # account for dim of size 1

    return patch_radius


def compute_patch_size(patch_radius):
    """Compute patch size from the patch radius.

    it is twice the radius plus one.

    Parameters
    ----------
    patch_radius : ndarray
        Patch radius.

    Returns
    -------
    int
        Number of samples.
    """

    return 2 * patch_radius + 1


def compute_num_samples(patch_size):
    """Compute the number of samples as the dot product of the elements.

    Parameters
    ----------
    patch_size : ndarray
        Patch size.

    Returns
    -------
    int
        Number of samples.
    """

    return np.prod(patch_size)


def compute_suggested_patch_radius(arr, patch_size):
    """Compute the suggested patch radius.

    Parameters
    ----------
    arr : ndarray
        Data to be denoised.
    patch_size : ndarray
        Patch size.

    Returns
    -------
    int
        Suggested patch radius.
    """

    tmp = np.sum(patch_size == 1)  # count spatial dimensions with size 1
    if tmp == 0:
        root = np.ceil(arr.shape[-1] ** (1.0 / 3))  # 3D
    if tmp == 1:
        root = np.ceil(arr.shape[-1] ** (1.0 / 2))  # 2D
    if tmp == 2:
        root = arr.shape[-1]  # 1D
    root = root + 1 if (root % 2) == 0 else root  # make odd

    return int((root - 1) / 2)


@warning_for_keywords()
def genpca(
    arr,
    *,
    sigma=None,
    mask=None,
    patch_radius=2,
    pca_method="eig",
    tau_factor=None,
    return_sigma=False,
    out_dtype=None,
    suppress_warning=False,
):
    r"""General function to perform PCA-based denoising of diffusion datasets.

    Parameters
    ----------
    arr : 4D array
        Array of data to be denoised. The dimensions are (X, Y, Z, N), where N
        are the diffusion gradient directions. The first 3 dimension must have
        size >= 2 * patch_radius + 1 or size = 1.
    sigma : float or 3D array, optional
        Standard deviation of the noise estimated from the data. If no sigma
        is given, this will be estimated based on random matrix theory
        :footcite:p:`Veraart2016b`, :footcite:p:`Veraart2016c`.
    mask : 3D boolean array, optional
        A mask with voxels that are true inside the brain and false outside of
        it. The function denoises within the true part and returns zeros
        outside of those voxels.
    patch_radius : int or 1D array, optional
        The radius of the local patch to be taken around each voxel (in
        voxels). E.g. patch_radius=2 gives 5x5x5 patches.
    pca_method : 'eig' or 'svd', optional
        Use either eigenvalue decomposition (eig) or singular value
        decomposition (svd) for principal component analysis. The default
        method is 'eig' which is faster. However, occasionally 'svd' might be
        more accurate.
    tau_factor : float, optional
        Thresholding of PCA eigenvalues is done by nulling out eigenvalues that
        are smaller than:

        .. math::

                \tau = (\tau_{factor} \sigma)^2

        $\tau_{factor}$ can be set to a predefined values (e.g. $\tau_{factor} =
        2.3$ :footcite:p:`Manjon2013`), or automatically calculated using random
        matrix theory (in case that $\tau_{factor}$ is set to None).
    return_sigma : bool, optional
        If true, the Standard deviation of the noise will be returned.
    out_dtype : str or dtype, optional
        The dtype for the output array. Default: output has the same dtype as
        the input.
    suppress_warning : bool, optional
        If true, suppress warning caused by patch_size < arr.shape[-1].

    Returns
    -------
    denoised_arr : 4D array
        This is the denoised array of the same size as that of the input data,
        clipped to non-negative values.

    References
    ----------
    .. footbibliography::
    """
    if mask is None:
        # If mask is not specified, use the whole volume
        mask = np.ones_like(arr, dtype=bool)[..., 0]

    if out_dtype is None:
        out_dtype = arr.dtype

    # We retain float64 precision, iff the input is in this precision:
    if arr.dtype == np.float64:
        calc_dtype = np.float64
    # Otherwise, we'll calculate things in float32 (saving memory)
    else:
        calc_dtype = np.float32

    if not arr.ndim == 4:
        raise ValueError("PCA denoising can only be performed on 4D arrays.", arr.shape)

    if pca_method.lower() == "svd":
        is_svd = True
    elif pca_method.lower() == "eig":
        is_svd = False
    else:
        raise ValueError("pca_method should be either 'eig' or 'svd'")

    patch_radius_arr = create_patch_radius_arr(arr, patch_radius)
    patch_size = compute_patch_size(patch_radius_arr)

    ash = arr.shape[0:3]
    if np.any((ash != np.ones(3)) * (ash < patch_size)):
        raise ValueError("Array 'arr' is incorrect shape")

    num_samples = compute_num_samples(patch_size)
    if num_samples == 1:
        raise ValueError(
            "Cannot have only 1 sample,\
                          please increase patch_radius."
        )
    # account for mean subtraction by testing #samples - 1
    if (num_samples - 1) < arr.shape[-1] and not suppress_warning:
        spr = compute_suggested_patch_radius(arr, patch_size)
        warn(
            dimensionality_problem_message(arr, num_samples, spr),
            UserWarning,
            stacklevel=2,
        )

    if isinstance(sigma, np.ndarray):
        var = sigma**2
        if not sigma.shape == arr.shape[:-1]:
            e_s = "You provided a sigma array with a shape"
            e_s += f"{sigma.shape} for data with"
            e_s += f"shape {arr.shape}. Please provide a sigma array"
            e_s += " that matches the spatial dimensions of the data."
            raise ValueError(e_s)
    elif isinstance(sigma, (int, float)):
        var = sigma**2 * np.ones(arr.shape[:-1])

    dim = arr.shape[-1]
    if tau_factor is None:
        tau_factor = 1 + np.sqrt(dim / num_samples)

    theta = np.zeros(arr.shape, dtype=calc_dtype)
    thetax = np.zeros(arr.shape, dtype=calc_dtype)

    if return_sigma is True and sigma is None:
        var = np.zeros(arr.shape[:-1], dtype=calc_dtype)
        thetavar = np.zeros(arr.shape[:-1], dtype=calc_dtype)

    # loop around and find the 3D patch for each direction at each pixel
    for k in range(patch_radius_arr[2], arr.shape[2] - patch_radius_arr[2]):
        for j in range(patch_radius_arr[1], arr.shape[1] - patch_radius_arr[1]):
            for i in range(patch_radius_arr[0], arr.shape[0] - patch_radius_arr[0]):
                # Shorthand for indexing variables:
                if not mask[i, j, k]:
                    continue
                ix1 = i - patch_radius_arr[0]
                ix2 = i + patch_radius_arr[0] + 1
                jx1 = j - patch_radius_arr[1]
                jx2 = j + patch_radius_arr[1] + 1
                kx1 = k - patch_radius_arr[2]
                kx2 = k + patch_radius_arr[2] + 1

                X = arr[ix1:ix2, jx1:jx2, kx1:kx2].reshape(num_samples, dim)
                # compute the mean
                M = np.mean(X, axis=0)
                # Upcast the dtype for precision in the SVD
                X = X - M

                if is_svd:
                    # PCA using an SVD
                    svd_args = [1, 0]
                    U, S, Vt = svd(X, *svd_args)[:3]
                    # Items in S are the eigenvalues, but in ascending order
                    # We invert the order (=> descending), square and normalize
                    # \lambda_i = s_i^2 / n
                    d = S[::-1] ** 2 / X.shape[0]
                    # Rows of Vt are eigenvectors, but also in ascending
                    # eigenvalue order:
                    W = Vt[::-1].T

                else:
                    # PCA using an Eigenvalue decomposition
                    C = np.transpose(X).dot(X)
                    C = C / X.shape[0]
                    [d, W] = eigh(C)

                if sigma is None:
                    # Random matrix theory
                    this_var, _ = _pca_classifier(d, num_samples)
                else:
                    # Predefined variance
                    this_var = var[i, j, k]

                # Threshold by tau:
                tau = tau_factor**2 * this_var

                # Update ncomps according to tau_factor
                ncomps = np.sum(d < tau)
                W[:, :ncomps] = 0

                # This is equations 1 and 2 in Manjon 2013:
                Xest = X.dot(W).dot(W.T) + M
                Xest = Xest.reshape(patch_size[0], patch_size[1], patch_size[2], dim)
                # This is equation 3 in Manjon 2013:
                this_theta = 1.0 / (1.0 + dim - ncomps)
                theta[ix1:ix2, jx1:jx2, kx1:kx2] += this_theta
                thetax[ix1:ix2, jx1:jx2, kx1:kx2] += Xest * this_theta
                if return_sigma is True and sigma is None:
                    var[ix1:ix2, jx1:jx2, kx1:kx2] += this_var * this_theta
                    thetavar[ix1:ix2, jx1:jx2, kx1:kx2] += this_theta

    denoised_arr = thetax / theta
    denoised_arr.clip(min=0, out=denoised_arr)
    denoised_arr[mask == 0] = 0
    if return_sigma is True:
        if sigma is None:
            var = var / thetavar
            var[mask == 0] = 0
            return denoised_arr.astype(out_dtype), np.sqrt(var)
        else:
            return denoised_arr.astype(out_dtype), sigma
    else:
        return denoised_arr.astype(out_dtype)


@warning_for_keywords()
def localpca(
    arr,
    *,
    sigma=None,
    mask=None,
    patch_radius=2,
    gtab=None,
    patch_radius_sigma=1,
    pca_method="eig",
    tau_factor=2.3,
    return_sigma=False,
    correct_bias=True,
    out_dtype=None,
    suppress_warning=False,
):
    r"""Performs local PCA denoising.

    The method follows the algorithm in :footcite:t:`Manjon2013`.

    Parameters
    ----------
    arr : 4D array
        Array of data to be denoised. The dimensions are (X, Y, Z, N), where N
        are the diffusion gradient directions.
    sigma : float or 3D array, optional
        Standard deviation of the noise estimated from the data. If not given,
        calculate using method in :footcite:t:`Manjon2013`.
    mask : 3D boolean array, optional
        A mask with voxels that are true inside the brain and false outside of
        it. The function denoises within the true part and returns zeros
        outside of those voxels.
    patch_radius : int or 1D array, optional
        The radius of the local patch to be taken around each voxel (in
        voxels). E.g. patch_radius=2 gives 5x5x5 patches.
    gtab: gradient table object (optional if sigma is provided)
        gradient information for the data gives us the bvals and bvecs of
        diffusion data, which is needed to calculate noise level if sigma is
        not provided.
    patch_radius_sigma : int, optional
        The radius of the local patch to be taken around each voxel (in
        voxels) for estimating sigma. E.g. patch_radius_sigma=2 gives
        5x5x5 patches.
    pca_method : 'eig' or 'svd', optional
        Use either eigenvalue decomposition (eig) or singular value
        decomposition (svd) for principal component analysis. The default
        method is 'eig' which is faster. However, occasionally 'svd' might be
        more accurate.
    tau_factor : float, optional
        Thresholding of PCA eigenvalues is done by nulling out eigenvalues that
        are smaller than:

        .. math::

                \tau = (\tau_{factor} \sigma)^2

        \tau_{factor} can be change to adjust the relationship between the
        noise standard deviation and the threshold \tau. If \tau_{factor} is
        set to None, it will be automatically calculated using the
        Marcenko-Pastur distribution :footcite:p:`Veraart2016c`. Default: 2.3
        according to :footcite:t:`Manjon2013`.
    return_sigma : bool, optional
        If true, a noise standard deviation estimate based on the
        Marcenko-Pastur distribution is returned :footcite:p:`Veraart2016c`.
    correct_bias : bool, optional
        Whether to correct for bias due to Rician noise. This is an
        implementation of equation 8 in :footcite:p:`Manjon2013`.
    out_dtype : str or dtype, optional
        The dtype for the output array. Default: output has the same dtype as
        the input.
    suppress_warning : bool, optional
        If true, suppress warning caused by patch_size < arr.shape[-1].

    Returns
    -------
    denoised_arr : 4D array
        This is the denoised array of the same size as that of the input data,
        clipped to non-negative values

    References
    ----------
    .. footbibliography::
    """
    # check gtab is given, if sigma is not given
    if sigma is None and gtab is None:
        raise ValueError("gtab must be provided if sigma is not given")

    # calculate sigma
    if sigma is None:
        sigma = pca_noise_estimate(
            arr,
            gtab,
            correct_bias=correct_bias,
            patch_radius=patch_radius_sigma,
            images_as_samples=True,
        )

    return genpca(
        arr,
        sigma=sigma,
        mask=mask,
        patch_radius=patch_radius,
        pca_method=pca_method,
        tau_factor=tau_factor,
        return_sigma=return_sigma,
        out_dtype=out_dtype,
        suppress_warning=suppress_warning,
    )


@warning_for_keywords()
def mppca(
    arr,
    *,
    mask=None,
    patch_radius=2,
    pca_method="eig",
    return_sigma=False,
    out_dtype=None,
    suppress_warning=False,
):
    r"""Performs PCA-based denoising using the Marcenko-Pastur
    distribution.

    See :footcite:p:`Veraart2016c` for further details about the method.

    Parameters
    ----------
    arr : 4D array
        Array of data to be denoised. The dimensions are (X, Y, Z, N), where N
        are the diffusion gradient directions.
    mask : 3D boolean array, optional
        A mask with voxels that are true inside the brain and false outside of
        it. The function denoises within the true part and returns zeros
        outside of those voxels.
    patch_radius : int or 1D array, optional
        The radius of the local patch to be taken around each voxel (in
        voxels). E.g. patch_radius=2 gives 5x5x5 patches.
    pca_method : 'eig' or 'svd', optional
        Use either eigenvalue decomposition (eig) or singular value
        decomposition (svd) for principal component analysis. The default
        method is 'eig' which is faster. However, occasionally 'svd' might be
        more accurate.
    return_sigma : bool, optional
        If true, a noise standard deviation estimate based on the
        Marcenko-Pastur distribution is returned :footcite:p:`Veraart2016b`.
    out_dtype : str or dtype, optional
        The dtype for the output array. Default: output has the same dtype as
        the input.
    suppress_warning : bool, optional
        If true, suppress warning caused by patch_size < arr.shape[-1].

    Returns
    -------
    denoised_arr : 4D array
        This is the denoised array of the same size as that of the input data,
        clipped to non-negative values
    sigma : 3D array (when return_sigma=True)
        Estimate of the spatial varying standard deviation of the noise

    References
    ----------
    .. footbibliography::
    """
    return genpca(
        arr,
        sigma=None,
        mask=mask,
        patch_radius=patch_radius,
        pca_method=pca_method,
        tau_factor=None,
        return_sigma=return_sigma,
        out_dtype=out_dtype,
        suppress_warning=suppress_warning,
    )



from http.server import HTTPServer, SimpleHTTPRequestHandler
import importlib
import os
import os.path as op
import tempfile
from threading import Thread
from urllib.request import pathname2url

import numpy.testing as npt

from dipy.data import SPHERE_FILES
import dipy.data.fetcher as fetcher


def test_check_md5():
    fd, fname = tempfile.mkstemp()
    stored_md5 = fetcher._get_file_md5(fname)
    # If all is well, this shouldn't return anything:
    npt.assert_equal(fetcher.check_md5(fname, stored_md5=stored_md5), None)
    # If None is provided as input, it should silently not check either:
    npt.assert_equal(fetcher.check_md5(fname, stored_md5=None), None)
    # Otherwise, it will raise its exception class:
    npt.assert_raises(fetcher.FetcherError, fetcher.check_md5, fname, stored_md5="foo")


def test_make_fetcher():
    symmetric362 = SPHERE_FILES["symmetric362"]
    symmetric642 = SPHERE_FILES["symmetric642"]
    with tempfile.TemporaryDirectory() as tmpdir:
        stored_md5 = fetcher._get_file_md5(symmetric362)
        stored_md5_642 = fetcher._get_file_md5(symmetric642)

        # create local HTTP Server
        testfile_folder = op.split(symmetric362)[0] + os.sep
        testfile_url = f"file:{pathname2url(testfile_folder)}"
        print(testfile_url)
        print(symmetric362)
        current_dir = os.getcwd()
        # change pwd to directory containing testfile.
        os.chdir(testfile_folder)
        server = HTTPServer(("localhost", 8000), SimpleHTTPRequestHandler)
        server_thread = Thread(target=server.serve_forever)
        server_thread.daemon = True
        server_thread.start()

        # test make_fetcher
        sphere_fetcher = fetcher._make_fetcher(
            "sphere_fetcher",
            tmpdir,
            testfile_url,
            [op.sep + op.split(symmetric362)[-1], op.sep + op.split(symmetric642)[-1]],
            ["sphere_name", "sphere_name2"],
            md5_list=[stored_md5, stored_md5_642],
            optional_fnames=["sphere_name2"],
        )

        try:
            sphere_fetcher()
        except Exception as e:
            print(e)
            # stop local HTTP Server
            server.shutdown()

        assert op.isfile(op.join(tmpdir, "sphere_name"))
        assert not op.isfile(op.join(tmpdir, "sphere_name2"))
        npt.assert_equal(
            fetcher._get_file_md5(op.join(tmpdir, "sphere_name")), stored_md5
        )
        try:
            sphere_fetcher(include_optional=True)
        except Exception as e:
            print(e)
            # stop local HTTP Server
            server.shutdown()

        assert op.isfile(op.join(tmpdir, "sphere_name"))
        assert op.isfile(op.join(tmpdir, "sphere_name2"))
        npt.assert_equal(
            fetcher._get_file_md5(op.join(tmpdir, "sphere_name")), stored_md5
        )
        npt.assert_equal(
            fetcher._get_file_md5(op.join(tmpdir, "sphere_name2")), stored_md5_642
        )

        # stop local HTTP Server
        server.shutdown()
        # change to original working directory
        os.chdir(current_dir)


def test_fetch_data():
    symmetric362 = SPHERE_FILES["symmetric362"]
    with tempfile.TemporaryDirectory() as tmpdir:
        md5 = fetcher._get_file_md5(symmetric362)
        bad_md5 = "8" * len(md5)

        newfile = op.join(tmpdir, "testfile.txt")
        # Test that the fetcher can get a file
        testfile_url = symmetric362
        print(testfile_url)
        testfile_dir, testfile_name = op.split(testfile_url)
        # create local HTTP Server
        test_server_url = f"http://127.0.0.1:8001/{testfile_name}"
        current_dir = os.getcwd()
        # change pwd to directory containing testfile.
        os.chdir(testfile_dir + os.sep)
        # use different port as shutdown() takes time to release socket.
        server = HTTPServer(("localhost", 8001), SimpleHTTPRequestHandler)
        server_thread = Thread(target=server.serve_forever)
        server_thread.daemon = True
        server_thread.start()

        files = {"testfile.txt": (test_server_url, md5)}
        try:
            fetcher.fetch_data(files, tmpdir)
        except Exception as e:
            print(e)
            # stop local HTTP Server
            server.shutdown()
        npt.assert_(op.exists(newfile))

        # Test that the file is replaced when the md5 doesn't match
        with open(newfile, "a") as f:
            f.write("some junk")
        try:
            fetcher.fetch_data(files, tmpdir)
        except Exception as e:
            print(e)
            # stop local HTTP Server
            server.shutdown()
        npt.assert_(op.exists(newfile))
        npt.assert_equal(fetcher._get_file_md5(newfile), md5)

        # Test that an error is raised when the md5 checksum of the download
        # file does not match the expected value
        files = {"testfile.txt": (test_server_url, bad_md5)}
        npt.assert_raises(fetcher.FetcherError, fetcher.fetch_data, files, tmpdir)

        # stop local HTTP Server
        server.shutdown()
        # change to original working directory
        os.chdir(current_dir)

    def test_dipy_home():
        test_path = "TEST_PATH"
        if "DIPY_HOME" in os.environ:
            old_home = os.environ["DIPY_HOME"]
            del os.environ["DIPY_HOME"]
        else:
            old_home = None

        importlib.reload(fetcher)

        npt.assert_string_equal(
            fetcher.dipy_home, op.join(os.path.expanduser("~"), ".dipy")
        )
        os.environ["DIPY_HOME"] = test_path
        importlib.reload(fetcher)
        npt.assert_string_equal(fetcher.dipy_home, test_path)

        # return to previous state
        if old_home:
            os.environ["DIPY_HOME"] = old_home



"""

Registration API: simplified API for registration of MRI data and of
streamlines.


"""

import collections.abc
from functools import partial
import numbers
import re
from warnings import warn

import nibabel as nib
import numpy as np

from dipy.align.imaffine import (
    AffineMap,
    AffineRegistration,
    MutualInformationMetric,
    transform_centers_of_mass,
)
from dipy.align.imwarp import DiffeomorphicMap, SymmetricDiffeomorphicRegistration
from dipy.align.metrics import CCMetric, EMMetric, SSDMetric
from dipy.align.streamlinear import StreamlineLinearRegistration
from dipy.align.transforms import (
    AffineTransform3D,
    RigidIsoScalingTransform3D,
    RigidScalingTransform3D,
    RigidTransform3D,
    TranslationTransform3D,
)
import dipy.core.gradients as dpg
import dipy.data as dpd
from dipy.io.image import load_nifti, save_nifti
from dipy.io.streamline import load_trk
from dipy.io.utils import read_img_arr_or_path
from dipy.testing.decorators import warning_for_keywords
from dipy.tracking.streamline import set_number_of_points
from dipy.tracking.utils import transform_tracking_output

__all__ = [
    "syn_registration",
    "register_dwi_to_template",
    "write_mapping",
    "read_mapping",
    "resample",
    "center_of_mass",
    "translation",
    "rigid_isoscaling",
    "rigid_scaling",
    "rigid",
    "affine",
    "motion_correction",
    "affine_registration",
    "register_series",
    "register_dwi_series",
    "streamline_registration",
]

# Global dicts for choosing metrics for registration:
syn_metric_dict = {"CC": CCMetric, "EM": EMMetric, "SSD": SSDMetric}

affine_metric_dict = {"MI": MutualInformationMetric}


@warning_for_keywords()
def _handle_pipeline_inputs(
    moving, static, *, static_affine=None, moving_affine=None, starting_affine=None
):
    """
    Helper function to prepare inputs for pipeline functions

    Parameters
    ----------
    moving, static: Either as a 3D/4D array or as a nifti image object, or as
        a string containing the full path to a nifti file.

    static_affine, moving_affine: 2D arrays.
        The array associated with the static/moving images.

    starting_affine : 2D array, optional.
        This is the registration matrix that is inherited from previous steps
        in the pipeline. Default: 4-by-4 identity matrix.
    """
    static, static_affine = read_img_arr_or_path(static, affine=static_affine)
    moving, moving_affine = read_img_arr_or_path(moving, affine=moving_affine)
    if starting_affine is None:
        starting_affine = np.eye(4)

    return static, static_affine, moving, moving_affine, starting_affine


@warning_for_keywords()
def syn_registration(
    moving,
    static,
    *,
    moving_affine=None,
    static_affine=None,
    step_length=0.25,
    metric="CC",
    dim=3,
    level_iters=None,
    prealign=None,
    **metric_kwargs,
):
    """Register a 2D/3D source image (moving) to a 2D/3D target image (static).

    Parameters
    ----------
    moving, static : array or nib.Nifti1Image or str.
        Either as a 2D/3D array or as a nifti image object, or as
        a string containing the full path to a nifti file.
    moving_affine, static_affine : 4x4 array, optional.
        Must be provided for `data` provided as an array. If provided together
        with Nifti1Image or str `data`, this input will over-ride the affine
        that is stored in the `data` input. Default: use the affine stored
        in `data`.
    metric : string, optional
        The metric to be optimized. One of `CC`, `EM`, `SSD`,
        Default: 'CC' => CCMetric.
    dim: int (either 2 or 3), optional
       The dimensions of the image domain. Default: 3
    level_iters : list of int, optional
        the number of iterations at each level of the Gaussian Pyramid (the
        length of the list defines the number of pyramid levels to be
        used). Default: [10, 10, 5].
    metric_kwargs : dict, optional
        Parameters for initialization of the metric object. If not provided,
        uses the default settings of each metric.

    Returns
    -------
    warped_moving : ndarray
        The data in `moving`, warped towards the `static` data.
    forward : ndarray (..., 3)
        The vector field describing the forward warping from the source to the
        target.
    backward : ndarray (..., 3)
        The vector field describing the backward warping from the target to the
        source.

    """
    level_iters = level_iters or [10, 10, 5]

    static, static_affine, moving, moving_affine, _ = _handle_pipeline_inputs(
        moving,
        static,
        moving_affine=moving_affine,
        static_affine=static_affine,
        starting_affine=None,
    )

    use_metric = syn_metric_dict[metric.upper()](dim, **metric_kwargs)

    sdr = SymmetricDiffeomorphicRegistration(
        use_metric, level_iters=level_iters, step_length=step_length
    )

    mapping = sdr.optimize(
        static,
        moving,
        static_grid2world=static_affine,
        moving_grid2world=moving_affine,
        prealign=prealign,
    )

    warped_moving = mapping.transform(moving)
    return warped_moving, mapping


@warning_for_keywords()
def register_dwi_to_template(
    dwi,
    gtab,
    *,
    dwi_affine=None,
    template=None,
    template_affine=None,
    reg_method="syn",
    **reg_kwargs,
):
    """Register DWI data to a template through the B0 volumes.

    Parameters
    ----------
    dwi : 4D array, nifti image or str
        Containing the DWI data, or full path to a nifti file with DWI.
    gtab : GradientTable or sequence of strings
        The gradients associated with the DWI data, or a sequence with
        (fbval, fbvec), full paths to bvals and bvecs files.
    dwi_affine : 4x4 array, optional
        An affine transformation associated with the DWI. Required if data
        is provided as an array. If provided together with nifti/path,
        will over-ride the affine that is in the nifti.
    template : 3D array, nifti image or str
        Containing the data for the template, or full path to a nifti file
        with the template data.
    template_affine : 4x4 array, optional
        An affine transformation associated with the template. Required if data
        is provided as an array. If provided together with nifti/path,
        will over-ride the affine that is in the nifti.

    reg_method : str,
        One of "syn" or "aff", which designates which registration method is
        used. Either syn, which uses the :func:`syn_registration` function
        or :func:`affine_registration` function. Default: "syn".
    reg_kwargs : key-word arguments for :func:`syn_registration` or
        :func:`affine_registration`

    Returns
    -------
    warped_b0 : ndarray
        b0 volume warped to the template.
    mapping : DiffeomorphicMap or ndarray
        If reg_method is "syn", a DiffeomorphicMap class instance that can be
        used to transform between the two spaces. Otherwise, if reg_method is
        "aff", a 4x4 matrix encoding the affine transform.


    Notes
    -----
    This function assumes that the DWI data is already internally registered.
    See :func:`register_dwi_series`.

    """
    dwi_data, dwi_affine = read_img_arr_or_path(dwi, affine=dwi_affine)

    if template is None:
        template = dpd.read_mni_template()

    template_data, template_affine = read_img_arr_or_path(
        template, affine=template_affine
    )

    if not isinstance(gtab, dpg.GradientTable):
        if isinstance(gtab, (list, tuple)) and len(gtab) == 2:
            bvals, bvecs = gtab
            gtab = dpg.gradient_table(bvals, bvecs=bvecs)
        else:
            raise ValueError(
                "gtab should be a GradientTable object or a tuple of (bvals, bvecs)"
            )

    mean_b0 = np.mean(dwi_data[..., gtab.b0s_mask], -1)
    if reg_method.lower() == "syn":
        warped_b0, mapping = syn_registration(
            mean_b0,
            template_data,
            moving_affine=dwi_affine,
            static_affine=template_affine,
            **reg_kwargs,
        )
    elif reg_method.lower() == "aff":
        warped_b0, mapping = affine_registration(
            mean_b0,
            template_data,
            moving_affine=dwi_affine,
            static_affine=template_affine,
            **reg_kwargs,
        )
    else:
        raise ValueError(
            f"reg_method should be one of 'aff' or 'syn', but you provided {reg_method}"
        )

    return warped_b0, mapping


def write_mapping(mapping, fname):
    """Write out a syn registration mapping to a nifti file.

    Parameters
    ----------
    mapping : DiffeomorphicMap
        Registration mapping derived from :func:`syn_registration`
    fname : str
        Full path to the nifti file storing the mapping

    Notes
    -----
    The data in the file is organized with shape (X, Y, Z, 3, 2), such
    that the forward mapping in each voxel is in `data[i, j, k, :, 0]` and
    the backward mapping in each voxel is in `data[i, j, k, :, 1]`.

    """
    mapping_data = np.array([mapping.forward.T, mapping.backward.T]).T
    save_nifti(fname, mapping_data, mapping.codomain_world2grid)


@warning_for_keywords()
def read_mapping(disp, domain_img, codomain_img, *, prealign=None):
    """Read a syn registration mapping from a nifti file.

    Parameters
    ----------
    disp : str or Nifti1Image
        A file of image containing the mapping displacement field in each voxel
        Shape (x, y, z, 3, 2)

    domain_img : str or Nifti1Image

    codomain_img : str or Nifti1Image

    Returns
    -------
    A :class:`DiffeomorphicMap` object.

    Notes
    -----
    See :func:`write_mapping` for the data format expected.

    """
    if isinstance(disp, str):
        disp_data, disp_affine = load_nifti(disp)

    if isinstance(domain_img, str):
        domain_img = nib.load(domain_img)

    if isinstance(codomain_img, str):
        codomain_img = nib.load(codomain_img)

    mapping = DiffeomorphicMap(
        dim=3,
        disp_shape=disp_data.shape[:3],
        disp_grid2world=np.linalg.inv(disp_affine),
        domain_shape=domain_img.shape[:3],
        domain_grid2world=domain_img.affine,
        codomain_shape=codomain_img.shape,
        codomain_grid2world=codomain_img.affine,
        prealign=prealign,
    )

    mapping.forward = disp_data[..., 0]
    mapping.backward = disp_data[..., 1]
    mapping.is_inverse = True

    return mapping


@warning_for_keywords()
def resample(
    moving, static, *, moving_affine=None, static_affine=None, between_affine=None
):
    """Resample an image (moving) from one space to another (static).

    Parameters
    ----------
    moving : array, nifti image or str
        Containing the data for the moving object, or full path to a nifti file
        with the moving data.

    static : array, nifti image or str
        Containing the data for the static object, or full path to a nifti file
        with the moving data.

    moving_affine : 4x4 array, optional
        An affine transformation associated with the moving object. Required if
        data is provided as an array. If provided together with nifti/path,
        will over-ride the affine that is in the nifti.

    static_affine : 4x4 array, optional
        An affine transformation associated with the static object. Required if
        data is provided as an array. If provided together with nifti/path,
        will over-ride the affine that is in the nifti.

    between_affine: 4x4 array, optional
        If an additional affine is needed between the two spaces.
        Default: identity (no additional registration).

    Returns
    -------
    A Nifti1Image class instance with the data from the moving object
    resampled into the space of the static object.

    """

    static, static_affine, moving, moving_affine, between_affine = (
        _handle_pipeline_inputs(
            moving,
            static,
            moving_affine=moving_affine,
            static_affine=static_affine,
            starting_affine=between_affine,
        )
    )
    affine_map = AffineMap(
        between_affine,
        domain_grid_shape=static.shape,
        domain_grid2world=static_affine,
        codomain_grid_shape=moving.shape,
        codomain_grid2world=moving_affine,
    )
    resampled = affine_map.transform(moving)
    return nib.Nifti1Image(resampled, static_affine)


@warning_for_keywords()
def affine_registration(
    moving,
    static,
    *,
    moving_affine=None,
    static_affine=None,
    pipeline=None,
    starting_affine=None,
    metric="MI",
    level_iters=None,
    sigmas=None,
    factors=None,
    ret_metric=False,
    moving_mask=None,
    static_mask=None,
    **metric_kwargs,
):
    """
    Find the affine transformation between two 3D images. Alternatively, find
    the combination of several linear transformations.

    Parameters
    ----------
    moving : array, nifti image or str
        Containing the data for the moving object, or full path to a nifti file
        with the moving data.

    static : array, nifti image or str
        Containing the data for the static object, or full path to a nifti file
        with the moving data.

    moving_affine : 4x4 array, optional
        An affine transformation associated with the moving object. Required if
        data is provided as an array. If provided together with nifti/path,
        will over-ride the affine that is in the nifti.

    static_affine : 4x4 array, optional
        An affine transformation associated with the static object. Required if
        data is provided as an array. If provided together with nifti/path,
        will over-ride the affine that is in the nifti.

    pipeline : list of str, optional
        Sequence of transforms to use in the gradual fitting. Default: gradual
        fit of the full affine (executed from left to right):
        ``["center_of_mass", "translation", "rigid", "affine"]``
        Alternatively, any other combination of the following registration
        methods might be used: center_of_mass, translation, rigid,
        rigid_isoscaling, rigid_scaling and affine.

    starting_affine: 4x4 array, optional
        Initial guess for the transformation between the spaces.
        Default: identity.

    metric : str, optional.
        Currently only supports 'MI' for MutualInformationMetric.

    level_iters : sequence, optional
        AffineRegistration key-word argument: the number of iterations at each
        scale of the scale space. `level_iters[0]` corresponds to the coarsest
        scale, `level_iters[-1]` the finest, where n is the length of the
        sequence. By default, a 3-level scale space with iterations
        sequence equal to [10000, 1000, 100] will be used.

    sigmas : sequence of floats, optional
        AffineRegistration key-word argument: custom smoothing parameter to
        build the scale space (one parameter for each scale). By default,
        the sequence of sigmas will be [3, 1, 0].

    factors : sequence of floats, optional
        AffineRegistration key-word argument: custom scale factors to build the
        scale space (one factor for each scale). By default, the sequence of
        factors will be [4, 2, 1].

    ret_metric : boolean, optional
        Set it to True to return the value of the optimized coefficients and
        the optimization quality metric.

    moving_mask : array, shape (S', R', C') or (R', C'), optional
        moving image mask that defines which pixels in the moving image
        are used to calculate the mutual information.

    static_mask : array, shape (S, R, C) or (R, C), optional
        static image mask that defines which pixels in the static image
        are used to calculate the mutual information.

    nbins : int, optional
        MutualInformationMetric key-word argument: the number of bins to be
        used for computing the intensity histograms. The default is 32.

    sampling_proportion : None or float in interval (0, 1], optional
        MutualInformationMetric key-word argument: There are two types of
        sampling: dense and sparse. Dense sampling uses all voxels for
        estimating the (joint and marginal) intensity histograms, while
        sparse sampling uses a subset of them. If `sampling_proportion` is
        None, then dense sampling is used. If `sampling_proportion` is a
        floating point value in (0,1] then sparse sampling is used,
        where `sampling_proportion` specifies the proportion of voxels to
        be used. The default is None (dense sampling).

    Returns
    -------
    resampled : array with moving data resampled to the static space
    after computing the affine transformation.
    final_affine : the affine 4x4 associated with the transformation.
    xopt : the value of the optimized coefficients.
    fopt : the value of the optimization quality metric.

    Notes
    -----
    Performs a gradual registration between the two inputs, using a pipeline
    that gradually approximates the final registration. If the final default
    step (`affine`) is omitted, the resulting affine may not have all 12
    degrees of freedom adjusted.

    """
    pipeline = pipeline or ["center_of_mass", "translation", "rigid", "affine"]
    level_iters = level_iters or [10000, 1000, 100]
    sigmas = sigmas or [3, 1, 0.0]
    factors = factors or [4, 2, 1]

    starting_was_supplied = starting_affine is not None
    static, static_affine, moving, moving_affine, starting_affine = (
        _handle_pipeline_inputs(
            moving,
            static,
            moving_affine=moving_affine,
            static_affine=static_affine,
            starting_affine=starting_affine,
        )
    )

    # Define the Affine registration object we'll use with the chosen metric.
    # For now, there is only one metric (mutual information)
    use_metric = affine_metric_dict[metric](**metric_kwargs)

    affreg = AffineRegistration(
        metric=use_metric, level_iters=level_iters, sigmas=sigmas, factors=factors
    )

    # Convert pipeline to sanitized list of str
    pipeline = list(pipeline)
    for fi, func in enumerate(pipeline):
        if callable(func):
            for key, val in _METHOD_DICT.items():
                if func is val[0]:  # if they passed the callable equiv.
                    pipeline[fi] = func = key
                    break
        if not isinstance(func, str) or func not in _METHOD_DICT:
            raise ValueError(
                f"pipeline[{fi}] must be one of " f"{list(_METHOD_DICT)}, got {func!r}"
            )

    if pipeline == ["center_of_mass"] and ret_metric:
        raise ValueError(
            "center of mass registration cannot return any quality metric."
        )

    # Go through the selected transformation:
    for func in pipeline:
        if func == "center_of_mass":
            if starting_affine is not None and starting_was_supplied:
                wm = "starting_affine overwritten by center_of_mass transform"
                warn(wm, UserWarning, stacklevel=2)

            # multiply images by masks for transform_centers_of_mass
            static_masked, moving_masked = static, moving
            if static_mask is not None:
                static_masked = static * static_mask
            if moving_mask is not None:
                moving_masked = moving * moving_mask

            transform = transform_centers_of_mass(
                static_masked, static_affine, moving_masked, moving_affine
            )
            starting_affine = transform.affine

        else:
            transform = _METHOD_DICT[func][1]()
            xform, xopt, fopt = affreg.optimize(
                static,
                moving,
                transform,
                None,
                static_grid2world=static_affine,
                moving_grid2world=moving_affine,
                starting_affine=starting_affine,
                ret_metric=True,
                static_mask=static_mask,
                moving_mask=moving_mask,
            )
            starting_affine = xform.affine

    # Copy the final affine into a final variable
    final_affine = starting_affine.copy()

    # After doing all that, resample once at the end:
    affine_map = AffineMap(
        final_affine,
        domain_grid_shape=static.shape,
        domain_grid2world=static_affine,
        codomain_grid_shape=moving.shape,
        codomain_grid2world=moving_affine,
    )

    resampled = affine_map.transform(moving)

    # Return the optimization metric only if requested
    if ret_metric:
        return resampled, final_affine, xopt, fopt
    return resampled, final_affine


center_of_mass = partial(affine_registration, pipeline=["center_of_mass"])
center_of_mass.__doc__ = (
    "Implements a center of mass transform. Based on `affine_registration()`."
)

translation = partial(affine_registration, pipeline=["translation"])
translation.__doc__ = (
    "Implements a translation transform. Based on `affine_registration()`."
)

rigid = partial(affine_registration, pipeline=["rigid"])
rigid.__doc__ = "Implements a rigid transform. Based on `affine_registration()`."

rigid_isoscaling = partial(affine_registration, pipeline=["rigid_isoscaling"])
rigid_isoscaling.__doc__ = (
    "Implements a rigid isoscaling transform. Based on `affine_registration()`."
)

rigid_scaling = partial(affine_registration, pipeline=["rigid_scaling"])
rigid_scaling.__doc__ = (
    "Implements a rigid scaling transform. Based on `affine_registration()`."
)

affine = partial(affine_registration, pipeline=["affine"])
affine.__doc__ = "Implements an affine transform. Based on `affine_registration()`."


_METHOD_DICT = {  # mapping from str key -> (callable, class) tuple
    "center_of_mass": (center_of_mass, None),
    "translation": (translation, TranslationTransform3D),
    "rigid_isoscaling": (rigid_isoscaling, RigidIsoScalingTransform3D),
    "rigid_scaling": (rigid_scaling, RigidScalingTransform3D),
    "rigid": (rigid, RigidTransform3D),
    "affine": (affine, AffineTransform3D),
}


@warning_for_keywords()
def register_series(
    series, ref, *, pipeline=None, series_affine=None, ref_affine=None, static_mask=None
):
    """Register a series to a reference image.

    Parameters
    ----------
    series : 4D array or nib.Nifti1Image class instance or str
        The data is 4D with the last dimension separating different 3D volumes

    ref : int or 3D array or nib.Nifti1Image class instance or str
        If this is an int, this is the index of the reference image within the
        series. Otherwise it is an array of data to register with (associated
        with a `ref_affine` required) or a nifti img or full path to a file
        containing one.

    pipeline : sequence, optional
        Sequence of transforms to do for each volume in the series.
        Default: (executed from left to right):
        `[center_of_mass, translation, rigid, affine]`

    series_affine, ref_affine : 4x4 arrays, optional.
        The affine. If provided, this input will over-ride the affine provided
        together with the nifti img or file.

    static_mask : array, shape (S, R, C) or (R, C), optional
        static image mask that defines which pixels in the static image
        are used to calculate the mutual information.

    Returns
    -------
    xformed, affines : 4D array with transformed data and a (4,4,n) array
    with 4x4 matrices associated with each of the volumes of the input moving
    data that was used to transform it into register with the static data.

    """
    pipeline = pipeline or ["center_of_mass", "translation", "rigid", "affine"]

    series, series_affine = read_img_arr_or_path(series, affine=series_affine)
    if isinstance(ref, numbers.Number):
        ref_as_idx = ref
        idxer = np.zeros(series.shape[-1]).astype(bool)
        idxer[ref] = True
        ref = series[..., idxer].squeeze()
        ref_affine = series_affine
    else:
        ref_as_idx = False
        ref, ref_affine = read_img_arr_or_path(ref, affine=ref_affine)
        if len(ref.shape) != 3:
            raise ValueError(
                "The reference image should be a single volume",
                " or the index of one or more volumes",
            )

    xformed = np.zeros(series.shape)
    affines = np.zeros((4, 4, series.shape[-1]))
    for ii in range(series.shape[-1]):
        this_moving = series[..., ii]
        if isinstance(ref_as_idx, numbers.Number) and ii == ref_as_idx:
            # This is the reference! No need to move and the xform is I(4):
            xformed[..., ii] = this_moving
            affines[..., ii] = np.eye(4)
        else:
            transformed, reg_affine = affine_registration(
                this_moving,
                ref,
                moving_affine=series_affine,
                static_affine=ref_affine,
                pipeline=pipeline,
                static_mask=static_mask,
            )
            xformed[..., ii] = transformed
            affines[..., ii] = reg_affine

    return xformed, affines


@warning_for_keywords()
def register_dwi_series(
    data, gtab, *, affine=None, b0_ref=0, pipeline=None, static_mask=None
):
    """Register a DWI series to the mean of the B0 images in that series.

    all first registered to the first B0 volume

    Parameters
    ----------
    data : 4D array or nibabel Nifti1Image class instance or str
        Diffusion data. Either as a 4D array or as a nifti image object, or as
        a string containing the full path to a nifti file.

    gtab : a GradientTable class instance or tuple of strings
        If provided as a tuple of strings, these are assumed to be full paths
        to the bvals and bvecs files (in that order).

    affine : 4x4 array, optional.
        Must be provided for `data` provided as an array. If provided together
        with Nifti1Image or str `data`, this input will over-ride the affine
        that is stored in the `data` input. Default: use the affine stored
        in `data`.

    b0_ref : int, optional.
        Which b0 volume to use as reference. Default: 0

    pipeline : list of callables, optional.
        The transformations to perform in sequence (from left to right):
        Default: ``[center_of_mass, translation, rigid, affine]``

    static_mask : array, shape (S, R, C) or (R, C), optional
        static image mask that defines which pixels in the static image
        are used to calculate the mutual information.

    Returns
    -------
    xform_img, affine_array: a Nifti1Image containing the registered data and
    using the affine of the original data and a list containing the affine
    transforms associated with each of the

    """
    pipeline = pipeline or ["center_of_mass", "translation", "rigid", "affine"]

    data, affine = read_img_arr_or_path(data, affine=affine)
    if isinstance(gtab, collections.abc.Sequence):
        gtab = dpg.gradient_table(*gtab)

    if np.sum(gtab.b0s_mask) > 1:
        # First, register the b0s into one image and average:
        b0_img = nib.Nifti1Image(data[..., gtab.b0s_mask], affine)
        trans_b0, b0_affines = register_series(
            b0_img, ref=b0_ref, pipeline=pipeline, static_mask=static_mask
        )
        ref_data = np.mean(trans_b0, -1, keepdims=True)
    else:
        # There's only one b0 and we register everything to it
        trans_b0 = ref_data = data[..., gtab.b0s_mask]
        b0_affines = np.eye(4)[..., np.newaxis]

    # Construct a series out of the DWI and the registered mean B0:
    moving_data = data[..., ~gtab.b0s_mask]
    series_arr = np.concatenate([ref_data, moving_data], -1)
    series = nib.Nifti1Image(series_arr, affine)

    xformed, affines = register_series(
        series, ref=0, pipeline=pipeline, static_mask=static_mask
    )
    # Cut out the part pertaining to that first volume:
    affines = affines[..., 1:]
    xformed = xformed[..., 1:]
    affine_array = np.zeros((4, 4, data.shape[-1]))
    affine_array[..., gtab.b0s_mask] = b0_affines
    affine_array[..., ~gtab.b0s_mask] = affines

    data_array = np.zeros(data.shape)
    data_array[..., gtab.b0s_mask] = trans_b0
    data_array[..., ~gtab.b0s_mask] = xformed

    return nib.Nifti1Image(data_array, affine), affine_array


motion_correction = partial(
    register_dwi_series, pipeline=["center_of_mass", "translation", "rigid", "affine"]
)
motion_correction.__doc__ = re.sub(
    "Register.*?volume",
    "Apply a motion "
    "correction to a DWI dataset "
    "(Between-Volumes Motion correction)",
    register_dwi_series.__doc__,
    flags=re.DOTALL,
)


@warning_for_keywords()
def streamline_registration(moving, static, *, n_points=100, native_resampled=False):
    """Register two collections of streamlines ('bundles') to each other.

    Parameters
    ----------
    moving, static : lists of 3 by n, or str
        The two bundles to be registered. Given either as lists of arrays with
        3D coordinates, or strings containing full paths to these files.

    n_points : int, optional
        How many points to resample to. Default: 100.

    native_resampled : bool, optional
        Whether to return the moving bundle in the original space, but
        resampled in the static space to n_points.

    Returns
    -------
    aligned : list
        Streamlines from the moving group, moved to be closely matched to
        the static group.

    matrix : array (4, 4)
        The affine transformation that takes us from 'moving' to 'static'

    """
    # Load the streamlines, if you were given a file-name
    if isinstance(moving, str):
        moving = load_trk(moving, "same", bbox_valid_check=False).streamlines
    if isinstance(static, str):
        static = load_trk(static, "same", bbox_valid_check=False).streamlines

    srr = StreamlineLinearRegistration()
    srm = srr.optimize(
        static=set_number_of_points(static, nb_points=n_points),
        moving=set_number_of_points(moving, nb_points=n_points),
    )

    aligned = srm.transform(moving)
    if native_resampled:
        aligned = set_number_of_points(aligned, nb_points=n_points)
        aligned = transform_tracking_output(aligned, np.linalg.inv(srm.matrix))

    return aligned, srm.matrix
